"""
Utilities to localise the application.

author: Matthew Casey

&copy; [Digital Content Analysis Technology Ltd](https://www.d-cat.co.uk)
"""

import i18n


class Localise:
    """
    This class is used to set up all localisation for the SDK.
    """

    # Default and fallback locale.
    _DEFAULT_LOCALE = 'en_GB'
    _FALLBACK_LOCALE = 'en'

    @classmethod
    def setup(cls):
        """
        Sets up localisation so that the translations can be found and the defaults defined. Note that this assumes that the translations.py file has been built.
        """

        # Load in the string translations...
        import fusion_platform.translations

        # ...and set the configuration options.
        i18n.set('fallback', Localise._FALLBACK_LOCALE)
        i18n.set('error_on_missing_translation', True)
        i18n.set('error_on_missing_placeholder', True)
        i18n.set('error_on_missing_plural', True)
        i18n.set('enable_memoization', True)

    @classmethod
    def set_locale(cls, user_locale):
        """
        Sets the current locale.

        Args:
            user_locale: The locale to be used.
        """
        i18n.set('locale', user_locale)



"""
Documentation generator file.

author: Matthew Casey

&copy; [Digital Content Analysis Technology Ltd](https://www.d-cat.co.uk)

Use this script to build the SDK HTML documentation.

Usage:
  python documentation.py <path_to_engine_translations>

where

- `path_to_engine_translations` is the path to the engine "translations.py" file.
"""

import i18n
import importlib.util
from marshmallow import fields
import os
import sys

from fusion_platform.common.utilities import string_camel_to_underscore
from fusion_platform.models.credit import CreditSchema
from fusion_platform.models.data import DataSchema
from fusion_platform.models.data_file import DataFileSchema
from fusion_platform.models.model import Model
from fusion_platform.models.organisation import OrganisationSchema
from fusion_platform.models.process import ProcessSchema
from fusion_platform.models.process_execution import ProcessExecutionSchema
from fusion_platform.models.process_service_execution import ProcessServiceExecutionSchema
from fusion_platform.models.process_service_execution_log import ProcessServiceExecutionLogSchema
from fusion_platform.models.service import ServiceSchema
from fusion_platform.models.user import UserSchema

# Check the usage.
if len(sys.argv) <= 1:
    print("Usage: python documentation.py <path_to_engine_translations>")
    exit(1)

# Load in the translations file from the engine so that we can extract the model field descriptions.
spec = importlib.util.spec_from_file_location("translations", sys.argv[1])
translations = importlib.util.module_from_spec(spec)
spec.loader.exec_module(translations)


def get_schema_name(schema):
    """
    Converts a schema object's name into a more friendly name.

    Args:
        schema: The schema object.

    Returns:
        The friendly name.
    """
    return schema.__class__.__name__.replace('Schema', '')


def output_schema_fields(schema, level=0, file=None):
    """
    Recursive method to output the documentation for a particular schema (or nested schema). This will write a Markdown file for a top-level schema.

    Args:
        schema: The schema to traverse.
        level: The nesting level. 0 indicates a top-level schema.
        file: The optional file used to write the documentation to.
    """
    schema_name = string_camel_to_underscore(get_schema_name(schema))

    # Create the documentation file for the top-level schema.
    if level == 0:
        filename = os.path.join('fusion_platform', 'models', f"{schema_name}.md")
        print(f"Creating {filename}")
        file = open(filename, 'w')

    # Step through every field in the schema and output its documentation to the file.
    for field in schema.fields:
        # Ignore any hidden fields.
        if Model._METADATA_HIDE not in schema.fields[field].metadata:
            # Get the description for the field. This comes from the engine translations which have been loaded, or the metadata for any pseudo-field.
            default = schema.fields[field].metadata.get('description', '')
            description = i18n.t(f"models.{schema_name.replace('_', '.', 1)}.{field}.description", default=default)

            # Output the field to the file.
            file.write(f"{'    ' * level}* **{field}**: {description}\n")

            # If the field has a nested schema, then recursively step through its fields.
            if hasattr(schema.fields[field], 'inner') and isinstance(schema.fields[field].inner, fields.Nested) and hasattr(schema.fields[field].inner, 'nested'):
                nested = schema.fields[field].inner.nested
                output_schema_fields(nested, level + 1, file)

    # Make sure we close the file at the end.
    if (level == 0) and (file is not None):
        file.close()


# Output the documentation for all schemas.
schemas = [CreditSchema(), DataSchema(), DataFileSchema(), OrganisationSchema(), ProcessSchema(), ProcessExecutionSchema(), ProcessServiceExecutionSchema(),
           ProcessServiceExecutionLogSchema(), ServiceSchema(), UserSchema()]

for schema in schemas:
    output_schema_fields(schema)



"""
Credit model class file.

author: Matthew Casey

&copy; [Digital Content Analysis Technology Ltd](https://www.d-cat.co.uk)
"""

from marshmallow import Schema, EXCLUDE

from fusion_platform.models import fields
from fusion_platform.models.model import Model


# Define the model schema classes. These are maintained from the API definitions.

class CreditMonthlySpendSchema(Schema):
    """
    Nested schema class for spend.
    """
    month = fields.DateTime(required=True)
    credits = fields.Decimal(required=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class CreditSsdsSchema(Schema):
    """
    Nested schema class for runtime SSDs.
    """
    ssds = fields.List(fields.UUID(required=True), required=True)
    credits = fields.Decimal(required=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class CreditSchema(Schema):
    """
    Schema class for credit model.

    Each credit model has the following fields (and nested fields):

    .. include::credit.md
    """
    id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    created_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    updated_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    organisation_id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    # Removed lock.

    any_credits = fields.Decimal(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    cloud_storage_credits = fields.Decimal(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    registry_storage_credits = fields.Decimal(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    runtime_any_credits = fields.Decimal(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    runtime_ssds = fields.List(fields.Nested(CreditSsdsSchema()), allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    spend = fields.List(fields.Nested(CreditMonthlySpendSchema()), allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class Credit(Model):
    """
    Credit model class providing attributes and methods to manipulate credit item details.
    """

    # Override the schema.
    _SCHEMA = CreditSchema()

    # Override the base model class name.
    _BASE_MODEL_CLASS_NAME = 'Organisation'  # A string to prevent circular imports.

    # Base path.
    _PATH_BASE = '/organisations/{organisation_id}/credits/{credit_id}'

    # Override the standard model paths.
    _PATH_GET = _PATH_BASE



"""
User model class file.

author: Matthew Casey

&copy; [Digital Content Analysis Technology Ltd](https://www.d-cat.co.uk)
"""

from marshmallow import Schema, EXCLUDE

from fusion_platform.models import fields
from fusion_platform.models.model import Model
from fusion_platform.models.organisation import Organisation
from fusion_platform.session import Session


# Define the model schema classes. These are maintained from the API definitions.

class UserOrganisationSchema(Schema):
    """
    Nested schema class for organisation.
    """

    id = fields.UUID(required=True)
    last = fields.Boolean(required=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class UserSchema(Schema):
    """
    Schema class for user model.

    Each user model has the following fields (and nested fields):

    .. include::user.md
    """
    id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    created_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    updated_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    email = fields.Email(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    # Removed email_verified.
    given_name = fields.String(required=True)
    family_name = fields.String(required=True)
    telephone = fields.String(allow_none=True)

    role = fields.Integer(allow_none=True)  # Changed to optional.

    # Removed location_ip_address.
    # Removed location_ip_country.

    default_output_storage_period = fields.Integer(required=True)

    time_zone = fields.String(required=True)
    locale = fields.String(required=True)
    units = fields.String(required=True)

    notification_service_status = fields.List(fields.String(required=True), allow_none=True)
    notification_account = fields.List(fields.String(required=True), allow_none=True)
    notification_processing_chain = fields.List(fields.String(required=True), allow_none=True)
    notification_new_features = fields.List(fields.String(required=True), allow_none=True)
    notification_contact = fields.List(fields.String(required=True), allow_none=True)

    organisations = fields.List(fields.Nested(UserOrganisationSchema()), allow_none=True, metadata={'hide': True})  # Changed to hide as an attribute.

    # Removed various flags.

    last_request_at = fields.DateTime(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    # Removed search.

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class User(Model):
    """
    User model class providing attributes and methods to manipulate user details.
    """

    # Override the schema.
    _SCHEMA = UserSchema()

    # Base path.
    _PATH_BASE = '/users/{user_id}'

    # Override the standard model paths.
    _PATH_DELETE = _PATH_BASE
    _PATH_GET = _PATH_BASE
    _PATH_PATCH = _PATH_BASE

    # Add in the custom model paths.
    _PATH_CHANGE_PASSWORD = f"{_PATH_BASE}/change_password"

    def change_password(self, old, new):
        """
        Changes the user password. The new password must conform to the current password policy.

        Args:
            old: The old password.
            new: The new password.

        Raises:
            RequestError: if the update fails.
        """
        body = {self.__class__.__name__: {'old_password': old, 'new_password': new}}
        self._session.request(path=self._get_path(self.__class__._PATH_CHANGE_PASSWORD), method=Session.METHOD_POST, body=body)

    def find_organisations(self, id=None, name=None):
        """
        Searches for organisations with the specified id and/or (non-unique) name, returning the first object found and an iterator.

        Args:
            id: The organisation id to search for.
            name: The name to search for (case-sensitive).

        Returns:
            The first found organisation object, or None if not found, and an iterator through the found organisation objects.

        Raises:
            RequestError: if any get fails.
            ModelError: if a model could not be loaded or validated from the Fusion Platform<sup>&reg;</sup>.
        """
        # Search for the id and/or name.
        organisation = None

        for item in self.organisations:
            if ((id is not None) and (str(id).lower() == str(item.id).lower())) or ((name is not None) and (item.name.lower().startswith(name.lower()))):
                organisation = item
                break

        return organisation, self.organisations

    @property
    def organisations(self):
        """
        Provides an iterator through the organisations which the user belongs to.

        Returns:
            An iterator through the organisations.

        Raises:
            RequestError: if any get fails.
            ModelError: if a model could not be loaded or validated from the Fusion Platform<sup>&reg;</sup>.
        """
        return Organisation._models_from_api_ids(self._session, [{self.__class__._FIELD_ID: organisation.get(self.__class__._FIELD_ID)} for organisation in
                                                                 self._model.get('organisations', [])])



"""
Data file model class file.

author: Matthew Casey

&copy; [Digital Content Analysis Technology Ltd](https://www.d-cat.co.uk)
"""

import i18n
from marshmallow import Schema, EXCLUDE

from fusion_platform.common.raise_thread import RaiseThread
from fusion_platform.common.utilities import dict_nested_get
from fusion_platform.models import fields
from fusion_platform.models.model import Model, ModelError


# Define the model schema classes. These are maintained from the API definitions.

class DataFileSelectorSchema(Schema):
    """
    Nested schema class for selector, category, data type, unit, format and statistics.
    """
    selector = fields.String(required=True)
    category = fields.String(allow_none=True)
    data_type = fields.String(allow_none=True)
    unit = fields.String(allow_none=True)
    validation = fields.String(allow_none=True)

    area = fields.Decimal(allow_none=True)
    length = fields.Decimal(allow_none=True)
    points = fields.Integer(allow_none=True)

    initial_values = fields.List(fields.String(required=True), allow_none=True)
    minimum = fields.Decimal(allow_none=True)
    maximum = fields.Decimal(allow_none=True)
    mean = fields.Decimal(allow_none=True)
    sd = fields.Decimal(allow_none=True)

    histogram_minimum = fields.Decimal(allow_none=True)
    histogram_maximum = fields.Decimal(allow_none=True)
    histogram = fields.List(fields.Decimal(required=True), allow_none=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class DataFileSchema(Schema):
    """
    Schema class for data file model.

    Each data file model has the following fields (and nested fields):

    .. include::data_file.md
    """
    id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    created_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    updated_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    organisation_id = fields.UUID(allow_none=True, metadata={'read_only': True, 'title': i18n.t('models.data_file.organisation_id.title'), 'description': i18n.t(
        'models.data_file.organisation_id.description')})  # Added parameter so that it can be used.
    data_id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    file_id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    preview_file_id = fields.UUID(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    file_type = fields.String(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    file_name = fields.String(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    resolution = fields.Integer(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    crs = fields.String(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    bounds = fields.List(fields.Decimal(required=True), allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    area = fields.Decimal(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    length = fields.Decimal(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    points = fields.Integer(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    mgrs_cells = fields.List(fields.String(required=True), allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    sinusoidal_cells = fields.List(fields.Tuple((fields.Integer, fields.Integer), required=True), allow_none=True,
                                   metadata={'read_only': True})  # Changed to prevent this being updated.

    size = fields.Integer(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    error = fields.Boolean(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    publishable = fields.Boolean(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    alternative = fields.UUID(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    source = fields.UUID(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    selectors = fields.List(fields.Nested(DataFileSelectorSchema()), allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    downloads = fields.Integer(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    geojson = fields.String(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    title = fields.String(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    description = fields.String(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class DataFile(Model):
    """
    DataFile model class providing attributes and methods to manipulate data file details.
    """

    # Override the schema.
    _SCHEMA = DataFileSchema()

    # Override the base model class name.
    _BASE_MODEL_CLASS_NAME = 'Data'  # A string to prevent circular imports.

    # No standard model paths as data file objects cannot be retrieved directly from the API.

    # Add in the custom model paths.
    _PATH_DOWNLOAD_FILE = '/organisations/{organisation_id}/data/{data_id}/download_file/{file_id}'

    # Plus query parameters.
    _QUERY_PREVIEW = 'preview'

    def __init__(self, session):
        """
        Initialises the object.

        Args:
            session: The linked session object for interfacing with the Fusion Platform<sup>&reg;</sup>.
        """
        super(DataFile, self).__init__(session)

        # Initialise the fields.
        self.__download_progress = None
        self.__download_thread = None

    def download(self, path, preview=False, wait=False):
        """
        Downloads the file to the specified path. Optionally waits for the download to complete.

        Args:
            path: The local path to download the file to.
            preview: Optionally specify that the preview (PNG) of the file should be downloaded instead of the main file.
            wait: Optionally wait for the download to complete? Default False.

        Raises:
            RequestError: if the download fails.
        """
        # Make sure no download is currently in progress.
        if self.__download_thread is not None:
            raise ModelError(i18n.t('models.data_file.download_already_in_progress'))

        # Obtain the download URL.
        url = self.download_url(preview=preview)

        # Start the download in a separate thread.
        self.__download_progress = (url, path, 0)
        self.__download_thread = RaiseThread(target=self._session.download_file, args=(url, path, self.__download_callback))
        self.__download_thread.start()

        # Optionally wait for completion.
        self.download_complete(wait=wait)  # Ignore response.

    def __download_callback(self, url, destination, size):
        """
        Callback method used to receive progress from download. Updates the download progress.

        Args:
            url: The URL to download as a file.
            destination: The destination file path.
            size: The total size in bytes so far downloaded.
        """
        self.__download_progress = (url, destination, size)

    def download_complete(self, wait=False):
        """
        Checks whether the download has completed. If an error has occurred during the download, then this will raise a corresponding exception. Optionally waits
        for the download to complete.

        Args:
            wait: Optionally wait for the download to complete? Default False.

        Returns:
            True if the download is complete.

        Raises:
            RequestError: if any request fails.
            ModelError: if no download is in progress.
        """
        # Make sure a download is in progress.
        if self.__download_thread is None:
            raise ModelError(i18n.t('models.data_file.no_download'))

        # Check the download thread. This will raise an exception if an error has occurred.
        finished = False

        try:
            # Join will return immediately because of the timeout, but will raise an exception if something has gone wrong.
            self.__download_thread.join(timeout=None if wait else 0)

            # Check if the thread is still running after the join.
            finished = not self.__download_thread.is_alive()

        except:
            # Something went wrong. Make sure we mark the download as finished and re-raise the error.
            finished = True
            raise

        finally:
            # Make sure we clear the progress and thread if it has finished.
            if finished:
                self.__download_progress = None
                self.__download_thread = None

        return finished

    def download_progress(self):
        """
        Returns current download progress.

        Returns:
            A tuple of the URL, destination and total number of bytes downloaded so far.

        Raises:
            ModelError: if no download is in progress.
        """
        # Make sure a download is in progress.
        if self.__download_thread is None:
            raise ModelError(i18n.t('models.data_file.no_download'))

        return self.__download_progress

    def download_url(self, preview=False):
        """
        Obtains a URL which can be used to download the file. This URL is only valid for up to 1 hour.

        Args:
            preview: Optionally specify that the preview (PNG) URL for the file should be obtained instead of the URL for the main file.

        Raises:
            RequestError: if the request fails.
        """
        response = self._session.request(path=self._get_path(self.__class__._PATH_DOWNLOAD_FILE, file_id=self.file_id, organisation_id=self.organisation_id),
                                         query_parameters={DataFile._QUERY_PREVIEW: preview})

        # Assume that the URL held within the expected key within the resulting dictionary.
        url = dict_nested_get(response, [self.__class__._RESPONSE_KEY_EXTRAS, self.__class__._RESPONSE_KEY_URL])

        if url is None:
            raise ModelError(i18n.t('models.data_file.failed_download_url'))

        return url



"""
Organisation model class file.

author: Matthew Casey

&copy; [Digital Content Analysis Technology Ltd](https://www.d-cat.co.uk)
"""

from functools import partial
from marshmallow import Schema, EXCLUDE

from fusion_platform.models import fields
from fusion_platform.models.credit import Credit
from fusion_platform.models.data import Data
from fusion_platform.models.model import Model
from fusion_platform.models.process import Process
from fusion_platform.models.service import Service


# Define the model schema classes. These are maintained from the API definitions.

class OrganisationUserSchema(Schema):
    """
    Nested schema class for users.
    """
    id = fields.UUID(allow_none=True)
    email = fields.Email(allow_none=True)
    roles = fields.List(fields.String(required=True))

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class OrganisationSchema(Schema):
    """
    Schema class for organisation model.

    Each organisation model has the following fields (and nested fields):

    .. include::organisation.md
    """
    id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    created_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    updated_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    name = fields.String(required=True)
    address_line_1 = fields.String(required=True)
    address_line_2 = fields.String(allow_none=True)
    address_town_city = fields.String(required=True)
    address_post_zip_code = fields.String(required=True)
    address_country = fields.String(required=True)

    payment_customer = fields.String(allow_none=True)
    payment_valid = fields.Boolean(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated and optional.
    # Removed payment_last_checked.

    income_customer = fields.String(allow_none=True)
    income_valid = fields.Boolean(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated and optional.
    # Removed income_last_checked.

    income_tax_rate = fields.Decimal(allow_none=True)
    income_tax_reference = fields.String(allow_none=True)

    currency = fields.String(allow_none=True)  # Changed to optional.

    users = fields.List(fields.Nested(OrganisationUserSchema()), allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    agreed_licenses = fields.List(fields.UUID(required=True), allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    offers = fields.List(fields.UUID(required=True), allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    # Removed runtime_stop_charge.
    # Removed runtime_error_charge.

    # Removed audit_services.

    maximum_output_storage_period = fields.Integer(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    maximum_file_downloads = fields.Integer(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    # Removed search.

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class Organisation(Model):
    """
    Organisation model class providing attributes and methods to manipulate organisation details.
    """

    # Override the schema.
    _SCHEMA = OrganisationSchema()

    # Base path.
    _PATH_BASE = '/organisations/{organisation_id}'

    # Override the standard model paths.
    _PATH_DELETE = _PATH_BASE
    _PATH_GET = _PATH_BASE
    _PATH_PATCH = _PATH_BASE

    # Add in the custom model paths.
    _PATH_DATA = f"{_PATH_BASE}/data/uploaded"
    _PATH_OWN_SERVICES = f"{_PATH_BASE}/services"
    _PATH_PROCESSES = f"{_PATH_BASE}/processes"
    _PATH_SERVICES = f"{_PATH_BASE}/services/latest"

    def create_data(self, name, file_type, files, wait=False):
        """
        Creates a data object for the organisation and uploads the corresponding files. Optionally waits for the upload and analysis to complete.

        Args:
            name: The name of the data object.
            file_type: The type of files that the data object will hold.
            files: The list of file paths to be uploaded.
            wait: Optionally wait for the upload and analysis to complete? Default False.

        Returns:
            The created data object.

        Raises:
            RequestError: if the create fails.
            ModelError: if the model could not be created and validated by the Fusion Platform<sup>&reg;</sup>.
        """
        # Get a new template for the data model.
        data = Data(self._session)
        data._new(organisation_id=self.id)

        # Now attempt to create the data item.
        data._create(name, file_type, files, wait)

        return data

    @property
    def credit(self):
        """
        Returns:
            The credit model for the organisation.

        Raises:
            RequestError: if any get fails.
            ModelError: if a model could not be loaded or validated from the Fusion Platform<sup>&reg;</sup>.
        """
        return Credit._model_from_api_id(self._session, organisation_id=self.id, id=self.id)  # Credit model id is the same as the organisation's id.

    @property
    def data(self):
        """
        Provides an iterator through the organisation's uploaded data objects.

        Returns:
            An iterator through the data objects.

        Raises:
            RequestError: if any get fails.
            ModelError: if a model could not be loaded or validated from the Fusion Platform<sup>&reg;</sup>.
        """
        return Data._models_from_api_path(self._session, self._get_path(self.__class__._PATH_DATA))

    def find_data(self, id=None, name=None, search=None):
        """
        Searches for uploaded data objects with the specified id and/or (non-unique) name, returning the first object found and an iterator.

        Args:
            id: The data id to search for.
            name: The name to search for (case-sensitive).
            search: The term to search for (case-insensitive).

        Returns:
            The first found data object, or None if not found, and an iterator through the found data objects.

        Raises:
            RequestError: if any get fails.
            ModelError: if a model could not be loaded or validated from the Fusion Platform<sup>&reg;</sup>.
        """
        # Note that name is a key field, and hence we can only search using begins with.
        filter = self.__class__._build_filter(
            [(self.__class__._FIELD_ID, self.__class__._FILTER_MODIFIER_EQ, id), (self.__class__._FIELD_NAME, self.__class__._FILTER_MODIFIER_BEGINS_WITH, name)])

        # Build the partial find generator and execute it.
        find = partial(Data._models_from_api_path, self._session, self._get_path(self.__class__._PATH_DATA), filter=filter, search=search)
        return self.__class__._first_and_generator(find)

    def find_processes(self, id=None, name=None, search=None):
        """
        Searches for the organisation's processes with the specified id and/or (non-unique) name, returning the first object found and an iterator.

        Args:
            id: The process id to search for.
            name: The name to search for (case-sensitive).
            search: The term to search for (case-insensitive).

        Returns:
            The first found process object, or None if not found, and an iterator through the found process objects.

        Raises:
            RequestError: if any get fails.
            ModelError: if a model could not be loaded or validated from the Fusion Platform<sup>&reg;</sup>.
        """
        filter = self.__class__._build_filter(
            [(self.__class__._FIELD_ID, self.__class__._FILTER_MODIFIER_EQ, id), (self.__class__._FIELD_NAME, self.__class__._FILTER_MODIFIER_CONTAINS, name)])

        # Build the partial find generator and execute it.
        find = partial(Process._models_from_api_path, self._session, self._get_path(self.__class__._PATH_PROCESSES), filter=filter, search=search)
        return self.__class__._first_and_generator(find)

    def find_services(self, id=None, ssd_id=None, name=None, keyword=None, search=None):
        """
        Searches for services with the specified id, SSD id, (non-unique) name and/or keywords, returning the first object found and an iterator.

        Args:
            id: The service id to search for.
            ssd_id: The SSD id to search for.
            name: The name to search for (case-sensitive).
            keyword: The keyword to search for (case-sensitive).
            search: The term to search for (case-insensitive).

        Returns:
            The first found service object, or None if not found, and an iterator through the found service objects.

        Raises:
            RequestError: if any get fails.
            ModelError: if a model could not be loaded or validated from the Fusion Platform<sup>&reg;</sup>.
        """
        # Note that name is a key field, and hence we can only search using begins with.
        filter = self.__class__._build_filter(
            [(self.__class__._FIELD_ID, self.__class__._FILTER_MODIFIER_EQ, id), (self.__class__._FIELD_SSD_ID, self.__class__._FILTER_MODIFIER_EQ, ssd_id),
             (self.__class__._FIELD_NAME, self.__class__._FILTER_MODIFIER_BEGINS_WITH, name),
             (self.__class__._FIELD_KEYWORDS, self.__class__._FILTER_MODIFIER_CONTAINS, keyword)])

        # Build the partial find generator and execute it.
        find = partial(Service._models_from_api_path, self._session, self._get_path(self.__class__._PATH_SERVICES), filter=filter, search=search)
        return self.__class__._first_and_generator(find)

    def new_process(self, name, service):
        """
        Creates a new template process from the service object. This process is not persisted to the Fusion Platform<sup>&reg;</sup>.

        Args:
            name: The name of the process.
            service: The service for which the process is to be created.

        Returns:
            The new template process object.

        Raises:
            RequestError: if the new fails.
            ModelError: if the model could not be created and validated by the Fusion Platform<sup>&reg;</sup>.
        """
        # Get a new template for the process model using the service.
        process = Process(self._session)
        process._new(query_parameters={Model._get_id_name(Service.__name__): service.id}, organisation_id=self.id, name=name)

        return process

    @property
    def own_services(self):
        """
        Provides an iterator through the services owned by the organisation. This includes services which may not yet be approved.

        Returns:
            An iterator through the service objects.

        Raises:
            RequestError: if any get fails.
            ModelError: if a model could not be loaded or validated from the Fusion Platform<sup>&reg;</sup>.
        """
        return Service._models_from_api_path(self._session, self._get_path(self.__class__._PATH_OWN_SERVICES))

    @property
    def processes(self):
        """
        Provides an iterator through the organisation's processes.

        Returns:
            An iterator through the process objects.

        Raises:
            RequestError if any get fails.
            ModelError if a model could not be loaded or validated from the Fusion Platform<sup>&reg;</sup>.
        """
        return Process._models_from_api_path(self._session, self._get_path(self.__class__._PATH_PROCESSES))

    @property
    def services(self):
        """
        Provides an iterator through the services available to the organisation for execution.

        Returns:
            An iterator through the service objects.

        Raises:
            RequestError: if any get fails.
            ModelError: if a model could not be loaded or validated from the Fusion Platform<sup>&reg;</sup>.
        """
        return Service._models_from_api_path(self._session, self._get_path(self.__class__._PATH_SERVICES))



"""
Fields classes.

author: Matthew Casey

&copy; [Digital Content Analysis Technology Ltd](https://www.d-cat.co.uk)
"""

from datetime import datetime, timedelta, timezone
from dateutil.relativedelta import relativedelta
import i18n
import json
from marshmallow import fields
import typing


#
# Groups together all the bespoke field types used by models so that they can be serialised appropriately to JSON and so that the validation errors are localised.
# All fields descend from Marshmallow fields.
#

class Boolean(fields.Boolean):
    """
    Defines a Boolean.
    """

    # Localised validation errors.
    default_error_messages = {'invalid': i18n.t('models.fields.boolean.invalid')}


class DateTime(fields.DateTime):
    """
    Defines a datetime.
    """

    # Localised validation errors.
    default_error_messages = {'invalid': i18n.t('models.fields.datetime.invalid'), 'invalid_awareness': i18n.t('models.fields.datetime.invalid_awareness'),
                              'format': i18n.t('models.fields.datetime.format'), }

    def _serialize(self, value, attr, obj, **kwargs):
        """
        Overrides the serialization. This is because when a nested model is validated, it may be passed as a string already (from an update), instead of a datetime.

        Args:
            value: The value to serialize.
            attr: The attribute/key in `data` to deserialize.
            obj: The associated object.
            kwargs: Field-specific keyword arguments.

        Returns:
            The serialised datetime.
        """

        if isinstance(value, str):
            return value

        # Note that if the datetime does not have a timezone, we enforce UTC. See also utilities#json_default.
        value = value.astimezone(timezone.utc) if isinstance(value, datetime) and (value.tzinfo is None) else value
        return super(DateTime, self)._serialize(value, attr, obj, **kwargs)

    def _deserialize(self, value, attr, data, **kwargs):
        """
        Overrides the deserialization. This is because when a model is validated, it uses deserialization to validate the values. However,
        because the datetime is native, and not a string, the validation fails. We therefore pass the validation if the value is already a
        datetime.

        Args:
            value: The value to deserialize.
            attr: The attribute/key in `data` to deserialize.
            data: The raw input data passed to `Schema.load`.
            kwargs: Field-specific keyword arguments.

        Raises:
            ValidationError: If an invalid value is passed or if a required value is missing.
        """
        if not value:
            return None

        if value and isinstance(value, datetime):
            return value

        return super(DateTime, self)._deserialize(value, attr, data, **kwargs)


class Decimal(fields.Decimal):
    """
    Defines a decimal value.
    """

    # Localised validation errors.
    default_error_messages = {'invalid': i18n.t('models.fields.decimal.invalid'), 'too_large': i18n.t('models.fields.decimal.too_large'),
                              'special': i18n.t('models.fields.decimal.special')}


class Dict(fields.Dict):
    """
    Defines a dictionary.
    """

    # Localised validation errors.
    default_error_messages = {'invalid': i18n.t('models.fields.dict.invalid')}


class Email(fields.Email):
    """
    Defines an email.
    """

    # Localised validation errors.
    default_error_messages = {'invalid': i18n.t('models.fields.email.invalid')}


class Float(fields.Float):
    """
    Defines a float.
    """

    # Localised validation errors.
    default_error_messages = {'invalid': i18n.t('models.fields.float.invalid'), 'too_large': i18n.t('models.fields.float.too_large'),
                              'special': i18n.t('models.fields.float.special')}


class Integer(fields.Integer):
    """
    Defines an integer.
    """

    # Localised validation errors.
    default_error_messages = {'invalid': i18n.t('models.fields.integer.invalid'), 'too_large': i18n.t('models.fields.integer.too_large')}


class IP(fields.IP):
    """
    Defines an IP address.
    """

    # Localised validation errors.
    default_error_messages = {'invalid_ip': i18n.t('models.fields.ip.invalid_ip')}


class List(fields.List):
    """
    Defines a list of other fields.
    """

    # Localised validation errors.
    default_error_messages = {'invalid': i18n.t('models.fields.list.invalid')}


class Nested(fields.Nested):
    """
    Defines a nested field with its own schema.
    """

    # Localised validation errors.
    default_error_messages = {'type': i18n.t('models.fields.nested.type')}


class RelativeDelta(fields.Field):
    """
    A relative delta field. Represents a relative datetime.
    """

    # Localised validation errors.
    default_error_messages = {'invalid': i18n.t('models.fields.relativedelta.invalid')}

    def _deserialize(self, value, attr, data, **kwargs):
        """
        Deserializes the field.

        Args:
            value: The value of the field.
            attr: The attribute. Not used.
            data: The data. Not used.
            kwargs: The additional keyword arguments. Not used.

        Returns:
            The serialized value or None if the value is None.
        """

        if not value:
            return None

        if value and isinstance(value, relativedelta):
            return value

        if not isinstance(value, (str, bytes)):
            raise self.make_error('invalid')

        # Attempt to parse the string.
        try:
            dictionary = json.loads(value)
            years = dictionary.get('years', 0) if dictionary.get('years', 0) is not None else 0
            months = dictionary.get('months', 0) if dictionary.get('months', 0) is not None else 0
            days = dictionary.get('days', 0) if dictionary.get('days', 0) is not None else 0
            leapdays = dictionary.get('leapdays', 0) if dictionary.get('leapdays', 0) is not None else 0
            weeks = dictionary.get('weeks', 0) if dictionary.get('weeks', 0) is not None else 0
            hours = dictionary.get('hours', 0) if dictionary.get('hours', 0) is not None else 0
            minutes = dictionary.get('minutes', 0) if dictionary.get('minutes', 0) is not None else 0
            seconds = dictionary.get('seconds', 0) if dictionary.get('seconds', 0) is not None else 0
            microseconds = dictionary.get('microseconds', 0) if dictionary.get('microseconds', 0) is not None else 0

            return relativedelta(years=years, months=months, days=days, leapdays=leapdays, weeks=weeks, hours=hours, minutes=minutes, seconds=seconds,
                                 microseconds=microseconds, year=dictionary.get('year'), month=dictionary.get('month'), day=dictionary.get('day'),
                                 weekday=dictionary.get('weekday'), hour=dictionary.get('hour'), minute=dictionary.get('minute'), second=dictionary.get('second'),
                                 microsecond=dictionary.get('microsecond'))

        except Exception as error:
            raise self.make_error('invalid') from error

    def _serialize(self, value, attr, obj, **kwargs):
        """
        Serializes the field.

        Args:
            value: The value of the field.
            attr: The attribute. Not used.
            obj: The object. Not used.
            kwargs: The additional keyword arguments. Not used.

        Returns:
            The serialized value or None if the value is None.
        """

        if value is None:
            return None

        # Create a string representation of the relativedelta which can be parsed.
        dictionary = {'years': value.years, 'months': value.months, 'days': value.days, 'leapdays': value.leapdays, 'weeks': value.weeks, 'hours': value.hours,
                      'minutes': value.minutes, 'seconds': value.seconds, 'microseconds': value.microseconds, 'year': value.year, 'month': value.month,
                      'day': value.day, 'weekday': value.weekday, 'hour': value.hour, 'minute': value.minute, 'second': value.second,
                      'microsecond': value.microsecond}

        return json.dumps(dictionary)


class String(fields.String):
    """
    Defines a String.
    """

    # Localised validation errors.
    default_error_messages = {'invalid': i18n.t('models.fields.string.invalid'), 'invalid_utf8': i18n.t('models.fields.string.invalid_utf8')}

    def deserialize(self, value: typing.Any, attr: str = None, data: typing.Mapping[str, typing.Any] = None, **kwargs):
        """
        Deserializes a value. This method is overridden to replace empty strings with None when allowed.

        Args:
            value: The value to deserialize.
            attr: The attribute/key in `data` to deserialize.
            data: The raw input data passed to `Schema.load`.
            kwargs: Field-specific keyword arguments.

        Raises:
            ValidationError: If an invalid value is passed or if a required value is missing.
        """
        if (value is not None) and isinstance(value, str) and (len(value) <= 0) and self.allow_none:
            value = None

        return super(String, self).deserialize(value, attr, data, **kwargs)


class TimeDelta(fields.Field):
    """
    Defines a time delta. Note that we do not use the base TimeDelta class as it assumes seconds as a precision and does not allow any flexibility.
    """

    # Localised validation errors.
    default_error_messages = {'invalid': i18n.t('models.fields.timedelta.invalid')}

    def _deserialize(self, value, attr, data, **kwargs):
        """
        Deserializes the field.

        Args:
            value: The value of the field.
            attr: The attribute. Not used.
            data: The data. Not used.
            kwargs: The additional keyword arguments. Not used.

        Returns:
            The serialized value or None if the value is None.
        """

        if not value:
            return None

        if value and isinstance(value, timedelta):
            return value

        if not isinstance(value, (str, bytes)):
            raise self.make_error('invalid')

        # Attempt to parse the string.
        try:
            dictionary = json.loads(value)
            days = dictionary.get('days', 0) if dictionary.get('days', 0) is not None else 0
            seconds = dictionary.get('seconds', 0) if dictionary.get('seconds', 0) is not None else 0
            microseconds = dictionary.get('microseconds', 0) if dictionary.get('microseconds', 0) is not None else 0

            return timedelta(days=days, seconds=seconds, microseconds=microseconds)

        except Exception as error:
            raise self.make_error('invalid') from error

    def _serialize(self, value, attr, obj, **kwargs):
        """
        Serializes the field.

        Args:
            value: The value of the field.
            attr: The attribute. Not used.
            obj: The object. Not used.
            kwargs: The additional keyword arguments. Not used.

        Returns:
            The serialized value or None if the value is None.
        """

        if value is None:
            return None

        # Create a string representation of the timedelta which can be parsed.
        dictionary = {'days': value.days, 'seconds': value.seconds, 'microseconds': value.microseconds}

        return json.dumps(dictionary)


class Tuple(fields.Tuple):
    """
    Defines a tuple of other fields.
    """

    # Localised validation errors.
    default_error_messages = {'invalid': i18n.t('models.fields.tuple.invalid')}


class Url(fields.Url):
    """
    Defines a URL by enhancing the Marshmallow URL field.
    """
    default_error_messages = {'invalid': i18n.t('models.fields.url.invalid_url')}


class UUID(fields.UUID):
    """
    Defines a UUID by enhancing the Marshmallow UUID field.
    """

    # Localised validation errors.
    default_error_messages = {'invalid_uuid': i18n.t('models.fields.uuid.invalid_uuid')}



"""
Data model class file.

author: Matthew Casey

&copy; [Digital Content Analysis Technology Ltd](https://www.d-cat.co.uk)
"""

import i18n
from marshmallow import Schema, EXCLUDE
import os
from time import sleep

from fusion_platform.common.raise_thread import RaiseThread
from fusion_platform.common.utilities import dict_nested_get
from fusion_platform.models import fields
from fusion_platform.models.data_file import DataFile
from fusion_platform.models.model import Model, ModelError
from fusion_platform.session import Session


# Define the model schema classes. These are maintained from the API definitions.

class DataIngesterSchema(Schema):
    """
    Nested schema class for ingesters.
    """
    ingester_id = fields.UUID(required=True)
    dates = fields.List(fields.DateTime(required=True), required=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class DataSchema(Schema):
    """
    Schema class for data model.

    Each data has the following fields (and nested fields):

    .. include::data.md
    """
    id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    created_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    updated_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    organisation_id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    name = fields.String(required=True)

    # Removed lock.
    unlinked = fields.String(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    unfulfilled = fields.Boolean(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    ingester_availability = fields.List(fields.Nested(DataIngesterSchema()), allow_none=True,
                                        metadata={'read_only': True})  # Changed to prevent this being updated.

    # Removed link_id.
    # Removed link_model.
    # Removed link_hash.

    bounds = fields.List(fields.Decimal(required=True), allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    file_with_preview = fields.UUID(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    uploaded_organisation_id = fields.UUID(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    deletable = fields.String(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    # Removed creator.
    # Removed search.

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class Data(Model):
    """
    Data model class providing attributes and methods to manipulate data item details.
    """

    # Override the schema.
    _SCHEMA = DataSchema()

    # Override the base model class name.
    _BASE_MODEL_CLASS_NAME = 'Organisation'  # A string to prevent circular imports.

    # Base path.
    _PATH_ROOT = '/organisations/{organisation_id}/data'
    _PATH_BASE = f"{_PATH_ROOT}/{{data_id}}"

    # Override the standard model paths.
    _PATH_COPY = f"{_PATH_BASE}/copy"
    _PATH_CREATE = _PATH_ROOT
    _PATH_DELETE = _PATH_BASE
    _PATH_GET = _PATH_BASE
    _PATH_NEW = f"{_PATH_ROOT}/new"
    _PATH_PATCH = _PATH_BASE

    # Add in the custom model paths.
    _PATH_ADD_FILE = f"{_PATH_BASE}/add_file"
    _PATH_FILES = f"{_PATH_BASE}/files"

    # Response keys.
    _RESPONSE_KEY_FILE = 'file'

    def __init__(self, session):
        """
        Initialises the object.

        Args:
            session: The linked session object for interfacing with the Fusion Platform<sup>&reg;</sup>.
        """
        super(Data, self).__init__(session)

        # Initialise the fields.
        self.__upload_threads = {}

    def __add_file(self, file_type, file):
        """
        Attempts to add a file to the data object and then starts its upload. The file is uploaded using a thread.

        Args:
            file_type: The type of file to add.
            file: The path to the file to add.

        Raises:
            RequestError: if the add fails.
        """
        # Make sure the file exists.
        if not os.path.exists(file):
            raise ModelError(i18n.t('models.data.failed_add_missing_file', file=file))

        # Add the file to the data model.
        body = {self.__class__.__name__: {'name': os.path.basename(file), 'file_type': file_type}}
        response = self._session.request(path=self._get_path(self.__class__._PATH_ADD_FILE), method=Session.METHOD_POST, body=body)

        # Assume that the file id is held within the expected key within the resulting dictionary.
        file_id = dict_nested_get(response, [self.__class__._RESPONSE_KEY_EXTRAS, self.__class__._RESPONSE_KEY_FILE])

        if file_id is None:
            raise ModelError(i18n.t('models.data.failed_add_file_id'))

        # Assume that the URL held within the expected key within the resulting dictionary.
        url = dict_nested_get(response, [self.__class__._RESPONSE_KEY_EXTRAS, self.__class__._RESPONSE_KEY_URL])

        if url is None:
            raise ModelError(i18n.t('models.data.failed_add_file_url'))

        # Make sure the file is unique so that we can keep track of it.
        if file_id in self.__upload_threads:
            raise ModelError(i18n.t('models.data.failed_add_file_not_unique'))

        # Create and start a thread for the upload.
        thread = None

        try:
            thread = RaiseThread(target=self._session.upload_file, args=(url, file))
            thread.start()

        finally:
            # Make sure we record the thread so we can monitor it, even if it fails.
            self.__upload_threads[file_id] = thread

    def check_analysis_complete(self, wait=False):
        """
        Checks that the analysis of all files associated with this data object is complete. Optionally waits for the analysis to complete.

        Args:
            wait: Optionally wait for the analysis to complete? Default False.

        Returns:
            True if the analysis is complete for all files.

        Raises:
            RequestError: if any get fails.
            ModelError: if a model could not be loaded or validated from the Fusion Platform<sup>&reg;</sup>.
        """
        while True:
            complete = True

            # Load in each of the file models and check that every file has been uploaded, and has a publishable or error field to indicate that the analysis is
            # complete.
            for file in self.files:
                self._logger.debug('file %s: %s', file.file_name, file.attributes)

                has_fields = hasattr(file, self.__class__._FIELD_SIZE) and (
                        hasattr(file, self.__class__._FIELD_PUBLISHABLE) or hasattr(file, self.__class__._FIELD_ERROR))

                if not has_fields:
                    self._logger.debug('file %s not yet analysed', file.file_name)
                    complete = False
                    break

            # Break the loop if we are not waiting or are complete.
            if (not wait) or complete:
                break

            # We are waiting, so block for a short while.
            sleep(self.__class__._API_UPDATE_WAIT_PERIOD)

        return complete

    def copy(self, name):
        """
        Creates a data object as a copy of the current object for the same organisation and with the same files. The copy is treated as if it has been uploaded,
        even if the source data object was created by the platform and not uploaded.

        Args:
            name: The name of the copy.

        Returns:
            The created copy.

        Raises:
            RequestError: if the copy fails.
            ModelError: if the model could not be created and validated by the Fusion Platform<sup>&reg;</sup>.
        """
        # Copy the data model.
        body = {self.__class__.__name__: {'name': name}}
        response = self._session.request(path=self._get_path(self.__class__._PATH_COPY), method=Session.METHOD_POST, body=body)

        # Assume that the copy data id is held within the expected key within the resulting dictionary.
        copy_data_id = dict_nested_get(response, [self.__class__._RESPONSE_KEY_MODEL, self.__class__._FIELD_ID])

        if copy_data_id is None:
            raise ModelError(i18n.t('models.data.failed_copy_id'))

        # Return the copy.
        return self.__class__._model_from_api_id(self._session, organisation_id=self.organisation_id, id=copy_data_id)

    def _create(self, name, file_type, files, wait=False):
        """
        Attempts to create the model object with the given values. This assumes the model template has been loaded first using the _new method, and that the model
        is then created using a POST RESTful request. This assumes that the post body contains key names which include the name of the model class.

        This method is overridden to also upload the corresponding files and optionally wait for the upload and analysis to complete.

        Args:
        name: The name of the data object.
        file_type: The type of files that the data object will hold.
        files: The list of file paths to be uploaded.
        wait: Optionally wait for the upload and analysis to complete? Default False.

        Returns:
            The created data object.

        Raises:
            RequestError: if the create fails.
            ModelError: if the model could not be created and validated by the Fusion Platform<sup>&reg;</sup>.
        """
        # Use the super method to create the data item with the correct attributes. This will raise an exception if anything fails.
        super(Data, self)._create(name=name)

        # Add each of the files, assuming that each is of the same file type, and start its upload in a thread.
        try:
            for file in files:
                self.__add_file(file_type, file)

        finally:
            # Optionally wait for completion. We must complete this even if an exception has occurred because there may be multiple files. However, we only do this
            # if any threads were started.
            if len(self.__upload_threads) > 0:
                self.create_complete(wait=wait)  # Ignore response.

    def create_complete(self, wait=False):
        """
        Checks whether the data object file(s) upload and analysis have completed. If an error has occurred during the upload and analysis, then this will raise
        a corresponding exception. Optionally waits for the upload and analysis to complete.

        Note, a failure of one file upload will not stop the upload of other files. Therefore, if an exception is raised, further calls to this method are required
        until all upload and analysis (or errored) has completed and this method returns True.

        Args:
            wait: Optionally wait for the upload and analysis to complete? Default False.

        Returns:
            True if the upload and analysis are complete for all files.

        Raises:
            RequestError: if any request fails.
            ModelError: if the upload or analysis failed.
        """
        # Make sure a create is in progress.
        if len(self.__upload_threads) <= 0:
            raise ModelError(i18n.t('models.data.no_create'))

        # Check the upload threads. This will raise an exception if an error has occurred.
        first_error = None

        for file_id, thread in self.__upload_threads.items():
            if thread is not None:
                thread_finished = False

                try:
                    # Join will return immediately because of the timeout, but will raise an exception if something has gone wrong.
                    thread.join(timeout=None if wait else 0)

                    # Check if the thread is still running after the join.
                    thread_finished = not thread.is_alive()

                except Exception as e:
                    # Something went wrong. Make sure we mark the upload as finished.
                    thread_finished = True

                    # If we are waiting for everything to complete, then we must not re-raise the error here so that everything else can be completed first.
                    # Instead, we save the error off and raise it later.
                    if wait:
                        first_error = e if first_error is None else first_error
                    else:
                        raise

                finally:
                    # Make sure we clear the thread if it has finished.
                    if thread_finished:
                        self.__upload_threads[file_id] = None

        # Have all the uploads finished?
        uploads_finished = all([value is None for value in self.__upload_threads.values()])

        # Now check whether the analysis has completed. We do this in a loop so that we can wait until completion, but break out if we are not waiting. If an
        # error occurs, then we assume that the analysis is complete to prevent indefinitely waiting.
        complete = False

        try:
            if uploads_finished:
                complete = self.check_analysis_complete(wait=wait)

        except Exception as e:
            # An error occurred, make sure we treat this as complete to prevent an indefinite loop.
            complete = True

            # If we already have an error to raise, make sure we do not overwrite it.
            first_error = e if first_error is None else first_error

        finally:
            if complete:
                # Tidy up any finished threads. This will allow us to indicate that everything is complete.
                self.__upload_threads = {file_id: thread for file_id, thread in self.__upload_threads.items() if thread is not None}

        # If we encountered an error, make sure we raise it.
        if first_error is not None:
            raise first_error

        return len(self.__upload_threads) <= 0

    @property
    def files(self):
        """
        Provides an iterator through the data object's files.

        Returns:
            An iterator through the data object's files.

        Raises:
            RequestError: if any get fails.
            ModelError: if a model could not be loaded or validated from the Fusion Platform<sup>&reg;</sup>.
        """
        return DataFile._models_from_api_path(self._session, self._get_path(self.__class__._PATH_FILES), organisation_id=self.organisation_id)



"""
Raise thread class file.

author: Matthew Casey

&copy; [Digital Content Analysis Technology Ltd](https://www.d-cat.co.uk)
"""

from threading import Thread


class RaiseThread(Thread):
    """
    Custom thread class used to make sure exceptions are raised on join. Based on:
    https://www.geeksforgeeks.org/handling-a-threads-exception-in-the-caller-thread-in-python/
    """

    def __init__(self, *args, **kwargs):
        """
        Initialises the object.

        :param args: Positional arguments.
        :param kwargs: Keyword arguments.
        """
        super(RaiseThread, self).__init__(*args, **kwargs)

        # Make sure the error field is defined.
        self.__error = None

    def join(self, timeout=None):
        """
        Wait until the thread terminates or the timeout occurs.

        Args:
            timeout: The amount of time to wait until the thread has terminated. Using None will bock indefinitely.
        """
        super(RaiseThread, self).join(timeout)

        # Make sure any exception is re-raised in the calling thread.
        if self.__error is not None:
            raise self.__error

    def run(self):
        """
        Runs the thread operation, but catches any error so that it can be re-raised on join.
        """
        try:
            super(RaiseThread, self).run()
        except Exception as e:
            self.__error = e



"""
Process service execution model class file.

author: Matthew Casey

&copy; [Digital Content Analysis Technology Ltd](https://www.d-cat.co.uk)
"""

from marshmallow import Schema, EXCLUDE
import os

from fusion_platform.models import fields
from fusion_platform.models.data import Data
from fusion_platform.models.model import Model
from fusion_platform.models.process_service_execution_log import ProcessServiceExecutionLog


# Define the model schema classes. These are maintained from the API definitions.

class ProcessServiceExecutionActionValueSchema(Schema):
    """
    Nested schema class for service action value.
    """
    name = fields.String(required=True)
    required = fields.Boolean(required=True)
    data_type = fields.String(required=True)
    default = fields.String(allow_none=True)
    validation = fields.String(allow_none=True)

    constant = fields.Boolean(allow_none=True)

    ssd_id = fields.UUID(allow_none=True)
    output = fields.Integer(allow_none=True)
    selector = fields.String(allow_none=True)
    data_id = fields.UUID(allow_none=True)

    expression = fields.String(allow_none=True)
    value = fields.String(allow_none=True)
    url = fields.Url(allow_none=True)
    advanced = fields.Boolean(allow_none=True)
    title = fields.String(allow_none=True)
    description = fields.String(allow_none=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ProcessServiceExecutionActionSchema(Schema):
    """
    Nested schema class for service action.
    """
    name = fields.String(required=True)
    values = fields.List(fields.Nested(ProcessServiceExecutionActionValueSchema()), required=True)
    url = fields.Url(allow_none=True)
    title = fields.String(allow_none=True)
    description = fields.String(allow_none=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ProcessServiceExecutionActionsSchema(Schema):
    """
    Schema class used to isolate actions for controller responses.
    """
    actions = fields.List(fields.Nested(ProcessServiceExecutionActionSchema()), allow_none=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ProcessServiceExecutionMetricSchema(Schema):
    """
    Nested schema class for metrics which are recorded during execution.
    """
    date = fields.DateTime(allow_none=True)
    memory_total_bytes = fields.Integer(allow_none=True)
    memory_free_bytes = fields.Integer(allow_none=True)
    swap_total_bytes = fields.Integer(allow_none=True)
    swap_free_bytes = fields.Integer(allow_none=True)
    tmp_total_bytes = fields.Integer(allow_none=True)
    tmp_free_bytes = fields.Integer(allow_none=True)
    s3_transfer_bytes = fields.Integer(allow_none=True)
    external_transfer_bytes = fields.Integer(allow_none=True)
    comment = fields.String(allow_none=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ProcessServiceExecutionOptionSchema(Schema):
    """
    Nested schema class for options which are used during execution.
    """
    name = fields.String(required=True)
    value = fields.String(allow_none=True)
    data_type = fields.String(required=True)
    validation = fields.String(allow_none=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ProcessServiceExecutionSchema(Schema):
    """
    Schema class for process service execution model.

    Each process service execution model has the following fields (and nested fields):

    .. include::process_service_execution.md
    """
    id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    created_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    updated_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    organisation_id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    process_execution_id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    process_id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    service_id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    image_id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    name = fields.String(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    started_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    ended_at = fields.DateTime(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    runtime = fields.Integer(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    # Removed input_size.
    cpu = fields.Decimal(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    memory = fields.Decimal(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    actions = fields.List(fields.Nested(ProcessServiceExecutionActionSchema()), allow_none=True)
    options = fields.List(fields.Nested(ProcessServiceExecutionOptionSchema()), allow_none=True,
                          metadata={'read_only': True})  # Changed to prevent this being updated.
    inputs = fields.List(fields.UUID(required=True), allow_none=True, metadata={'hide': True})  # Changed to hide as an attribute.
    outputs = fields.List(fields.UUID(required=True), allow_none=True, metadata={'hide': True})  # Changed to hide as an attribute.
    intermediate = fields.Boolean(allow_none=True)
    # Removed task_id.
    # Removed task_definition_id.
    # Removed access_point_ids.
    # Removed security_group_id.
    success = fields.Boolean(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    metrics = fields.List(fields.Nested(ProcessServiceExecutionMetricSchema()), allow_none=True,
                          metadata={'read_only': True})  # Changed to prevent this being updated.

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ProcessServiceExecution(Model):
    """
    Process service execution model class providing attributes and methods to manipulate process execution item details.
    """

    # Override the schema.
    _SCHEMA = ProcessServiceExecutionSchema()

    # Override the base model class name.
    _BASE_MODEL_CLASS_NAME = 'Organisation'  # A string to prevent circular imports.

    # Base path.
    _PATH_BASE = '/organisations/{organisation_id}/process_service_executions/{process_service_execution_id}'

    # Override the standard model paths.
    _PATH_GET = _PATH_BASE

    # Add in the custom model paths.
    _PATH_LOGS = f"{_PATH_BASE}/logs"

    @property
    def inputs(self):
        """
        Provides an iterator through the process execution component's inputs.

        Returns:
            An iterator through the input data objects.

        Raises:
            RequestError: if any get fails.
            ModelError: if a model could not be loaded or validated from the Fusion Platform<sup>&reg;</sup>.
        """
        inputs = self._model.get(self.__class__._FIELD_INPUTS, []) if self._model.get(self.__class__._FIELD_INPUTS) is not None else []

        return Data._models_from_api_ids(self._session, Data._get_ids_from_list(inputs, organisation_id=self.organisation_id))

    def download_log_file(self, path):
        """
        Downloads the log output from the component's execution to the specified path as a CSV file.

        Args:
            path: The path to save the log records to.

        Raises:
            RequestError if any get fails.
        """
        logs = ProcessServiceExecutionLog._models_from_api_path(self._session, self._get_path(self.__class__._PATH_LOGS), reverse=True)
        first = True

        try:
            with open(path, 'w') as file:
                for log in logs:
                    header, line = log.to_csv(exclude=['id', 'created_at', 'updated_at', 'process_service_execution_id'])

                    if first:
                        file.write(header)
                        file.write('\n')

                        first = False

                    file.write(line)
                    file.write('\n')
        except:
            # If an error occurred, delete the file.
            os.remove(path)

            # Make sure the error is raised.
            raise

    @property
    def outputs(self):
        """
        Provides an iterator through the process execution component's outputs.

        Returns:
            An iterator through the output data objects.

        Raises:
            RequestError: if any get fails.
            ModelError: if a model could not be loaded or validated from the Fusion Platform<sup>&reg;</sup>.
        """
        outputs = self._model.get(self.__class__._FIELD_OUTPUTS, []) if self._model.get(self.__class__._FIELD_OUTPUTS) is not None else []

        return Data._models_from_api_ids(self._session, Data._get_ids_from_list(outputs, organisation_id=self.organisation_id))



"""
Translations generator file.

author: Matthew Casey

&copy; [Digital Content Analysis Technology Ltd](https://www.d-cat.co.uk)
"""

import os
from pathlib import Path
import yaml


def get_all_values(nested_dictionary, values=None, hierarchy=None, prefix=None, strip=None):
    """
    Recursively builds a dictionary of all the leaves in a nested dictionary.

    Args:
        nested_dictionary: The nested dictionary to traverse.
        values: The current set of extracted leaf values.
        hierarchy: The hierarchy for the keys.
        prefix: Any prefix to add to the keys.
        strip: Any value which should be stripped from the hierarchy.

    Returns:
        The dictionary of leaf values.
    """
    values = {} if values is None else values
    hierarchy = [] if hierarchy is None else hierarchy
    prefix = [] if prefix is None else prefix

    for key, value in nested_dictionary.items():
        path = hierarchy + [key]

        try:
            path.remove(strip)
        except:
            pass

        if type(value) is dict:
            values = {**get_all_values(value, hierarchy=path, prefix=prefix, strip=strip), **values}
        else:
            values = {'.'.join(prefix + path): value, **values}

    return values


# Find all the translation files, parse them and then build a complete list of the translations.
translations_by_locale = {}

for path in Path('resources/locales').rglob('*.yml'):
    print(f"Extracting {path}")
    locale = path.suffixes[0].replace('.', '')
    prefix = list(path.parts[2:-1])  # Strip off the 'resources/locales'.
    prefix += [os.path.splitext(path.stem)[0]]

    translations_by_locale[locale] = {} if translations_by_locale.get(locale) is None else translations_by_locale[locale]

    with open(path, 'r') as file:
        translations_by_locale[locale] = {**translations_by_locale[locale], **get_all_values(yaml.safe_load(file), prefix=prefix, strip=locale)}

# Now output all the translations as a Python file.
with open('fusion_platform/translations.py', 'w') as file:
    file.write('"""\n')
    file.write('Compiled translations.\n')
    file.write('\n')
    file.write('author: Matthew Casey\n')
    file.write('\n')
    file.write('&copy; [Digital Content Analysis Technology Ltd](https://www.d-cat.co.uk)\n')
    file.write('"""\n')
    file.write('\n')
    file.write('# Do not modify this file manually as it is built automatically by the localisations.py script.\n')
    file.write('\n')
    file.write('import i18n\n')
    file.write('\n')
    file.write('# @formatter:off\n')

    for locale, translations in translations_by_locale.items():
        for key, value in translations.items():
            translation = value.replace("'", "\\'")
            file.write(f"i18n.add_translation('{key}', '{translation}', '{locale}')\n")

    file.write('# @formatter:on\n')



"""
Service model class file.

author: Matthew Casey

&copy; [Digital Content Analysis Technology Ltd](https://www.d-cat.co.uk)
"""

from marshmallow import Schema, EXCLUDE

from fusion_platform.models import fields
from fusion_platform.models.model import Model


# Define the model schema classes. These are maintained from the API definitions.

class ServiceActionValueSchema(Schema):
    """
    Nested schema class for service action value.
    """
    name = fields.String(required=True)
    required = fields.Boolean(required=True)
    data_type = fields.String(required=True)
    default = fields.String(allow_none=True)
    validation = fields.String(allow_none=True)

    # Values can be:
    #  constant,
    #  calculated from an expression (using previously calculated values),
    #  taken from an SSD output,
    #  obtained from a URL, or
    #  user provided (with a default).

    constant = fields.Boolean(allow_none=True)

    ssd_id = fields.UUID(allow_none=True)
    output = fields.Integer(allow_none=True)
    selector = fields.String(allow_none=True)

    expression = fields.String(allow_none=True)

    value = fields.String(allow_none=True)
    url = fields.Url(allow_none=True)
    advanced = fields.Boolean(allow_none=True)
    title = fields.String(allow_none=True)
    description = fields.String(allow_none=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ServiceActionSchema(Schema):
    """
    Nested schema class for service action.
    """
    name = fields.String(required=True)
    values = fields.List(fields.Nested(ServiceActionValueSchema()), required=True)
    url = fields.Url(allow_none=True)
    title = fields.String(allow_none=True)
    description = fields.String(allow_none=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ServiceDefinitionLinkageSchema(Schema):
    """
    Nested schema class for service definition linkage.
    """
    ssd_id = fields.UUID(required=True)
    input = fields.Integer(required=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ServiceDefinitionSchema(Schema):
    """
    Nested schema class for service definition.
    """
    ssd_id = fields.UUID(required=True)
    output = fields.Integer(required=True)
    linkages = fields.List(fields.Nested(ServiceDefinitionLinkageSchema()), required=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ServiceGroupAggregatorOptionSchema(Schema):
    """
    Nested schema class for group aggregator option.
    """
    ssd_id = fields.UUID(required=True)
    name = fields.String(required=True)
    value = fields.String(required=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ServiceGroupAggregatorSchema(Schema):
    """
    Nested schema class for group aggregator.
    """
    aggregator_ssd_id = fields.UUID(required=True)
    output_ssd_id = fields.UUID(required=True)
    outputs = fields.List(fields.Integer(required=True), required=True)
    options = fields.List(fields.Nested(ServiceGroupAggregatorOptionSchema()), allow_none=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ServiceInputExpressionSchema(Schema):
    """
    Nested schema class for service input expression.
    """
    lhs_ssd_id = fields.UUID(required=True)
    lhs_input = fields.Integer(required=True)
    expression = fields.String(required=True)
    rhs_ssd_id = fields.UUID(required=True)
    rhs_input = fields.Integer(required=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ServiceOptionExpressionSchema(Schema):
    """
    Nested schema class for service option expression.
    """
    lhs_ssd_id = fields.UUID(required=True)
    lhs_name = fields.String(required=True)
    expression = fields.String(required=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ServiceValidationSchema(Schema):
    """
    Nested schema class for service validation.
    """
    expression = fields.String(required=True)
    message = fields.String(required=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ServiceSchema(Schema):
    """
    Schema class for service model.

    Each service model has the following fields (and nested fields):

    .. include::service.md
    """
    id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    created_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    updated_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    organisation_id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    ssd_id = fields.UUID(required=True)

    version = fields.Integer(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    approval_status = fields.String(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    # Removed latest.
    featured = fields.String(allow_none=True)
    show_in_latest = fields.Boolean(allow_none=True)  # Changed to optional.
    name = fields.String(required=True)

    categories = fields.List(fields.String(required=True), required=True)
    keywords = fields.List(fields.String(required=True), required=True)

    image_id = fields.UUID(required=True)

    definition = fields.List(fields.Nested(ServiceDefinitionSchema()), allow_none=True)
    group_aggregators = fields.List(fields.Nested(ServiceGroupAggregatorSchema()), allow_none=True)
    actions = fields.List(fields.Nested(ServiceActionSchema()), allow_none=True)
    urls = fields.List(fields.Url(required=True), allow_none=True)
    cidrs = fields.List(fields.String(required=True), allow_none=True)

    input_expressions = fields.List(fields.Nested(ServiceInputExpressionSchema()), allow_none=True)
    input_validations = fields.List(fields.Nested(ServiceValidationSchema()), allow_none=True)
    option_expressions = fields.List(fields.Nested(ServiceOptionExpressionSchema()), allow_none=True)
    option_validations = fields.List(fields.Nested(ServiceValidationSchema()), allow_none=True)

    license_id = fields.UUID(required=True)

    charge_expression_platform = fields.String(allow_none=True)  # Changed to optional.
    charge_expression_owner = fields.String(allow_none=True)  # Changed to optional.

    organisations = fields.List(fields.UUID(required=True), allow_none=True)
    geographic_regions = fields.List(fields.String(required=True), allow_none=True)

    # Removed creator.
    # Removed approver.
    # Removed search.

    documentation_summary = fields.String(required=True)
    documentation_description = fields.String(required=True)
    documentation_assumptions = fields.String(required=True)
    documentation_performance = fields.String(required=True)
    documentation_actions = fields.String(required=True)
    documentation_inputs = fields.List(fields.String(required=True), allow_none=True)  # Added pseudo-parameter.
    documentation_outputs = fields.List(fields.String(required=True), allow_none=True)  # Added pseudo-parameter.
    documentation_options = fields.List(fields.String(required=True), allow_none=True)  # Added pseudo-parameter.

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class Service(Model):
    """
    Service model class providing attributes and methods to manipulate service item details.
    """

    # Override the schema.
    _SCHEMA = ServiceSchema()

    # Override the base model class name.
    _BASE_MODEL_CLASS_NAME = 'Organisation'  # A string to prevent circular imports.

    # Base path.
    _PATH_BASE = '/organisations/{organisation_id}/services/{service_id}'

    # Override the standard model paths.
    _PATH_DELETE = _PATH_BASE
    _PATH_GET = _PATH_BASE
    _PATH_PATCH = _PATH_BASE



"""
Main entry point for the command line.

author: Matthew Casey

&copy; [Digital Content Analysis Technology Ltd](https://www.d-cat.co.uk)
"""

from fusion_platform.base import Base


def fusion_platform():
    """
    Main entry point for command line.
    """
    Base.main()



"""
Process execution model class file.

author: Matthew Casey

&copy; [Digital Content Analysis Technology Ltd](https://www.d-cat.co.uk)
"""

import i18n
from marshmallow import Schema, EXCLUDE
from time import sleep

from fusion_platform.models import fields
from fusion_platform.models.model import Model, ModelError
from fusion_platform.models.process_service_execution import ProcessServiceExecution


# Define the model schema classes. These are maintained from the API definitions.

class ProcessExecutionChainOptionSchema(Schema):
    """
    Nested schema class for SSD chain option.
    """
    name = fields.String(required=True)
    value = fields.String(allow_none=True)
    data_type = fields.String(required=True)
    validation = fields.String(allow_none=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ProcessExecutionChainSchema(Schema):
    """
    Nested schema class for SSD chain.
    """
    ssd_id = fields.UUID(required=True)
    service_id = fields.UUID(required=True)
    inputs = fields.List(fields.List(fields.UUID(allow_none=True), allow_none=True), allow_none=True)
    outputs = fields.List(fields.UUID(required=True), allow_none=True)
    options = fields.List(fields.Nested(ProcessExecutionChainOptionSchema()), allow_none=True)
    intermediate = fields.Boolean(allow_none=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ProcessExecutionOptionSchema(Schema):
    """
    Nested schema class for options which are provided to the SSD images when run.
    """
    ssd_id = fields.UUID(required=True)
    name = fields.String(required=True)
    value = fields.String(allow_none=True)
    required = fields.Boolean(required=True)
    data_type = fields.String(required=True)
    validation = fields.String(allow_none=True)
    mutually_exclusive = fields.String(allow_none=True)
    advanced = fields.Boolean(allow_none=True)
    title = fields.String(allow_none=True)
    description = fields.String(allow_none=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ProcessExecutionSchema(Schema):
    """
    Schema class for process execution model.

    Each process execution model has the following fields (and nested fields):

    .. include::process_execution.md
    """
    id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    created_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    updated_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    organisation_id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    process_id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    group_id = fields.UUID(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    group_index = fields.Integer(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    group_count = fields.Integer(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    options = fields.List(fields.Nested(ProcessExecutionOptionSchema()), allow_none=True)
    chains = fields.List(fields.Nested(ProcessExecutionChainSchema()), allow_none=True)
    # Removed creator.

    started_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    ended_at = fields.DateTime(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    # Removed wait_for_inputs.
    stopped = fields.Boolean(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    abort = fields.Boolean(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    abort_reason = fields.String(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    success = fields.Boolean(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    progress = fields.Integer(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    delete_expiry = fields.DateTime(required=True)
    delete_warning_status = fields.String(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    deletable = fields.String(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ProcessExecution(Model):
    """
    Process execution model class providing attributes and methods to manipulate process execution item details.
    """

    # Override the schema.
    _SCHEMA = ProcessExecutionSchema()

    # Override the base model class name.
    _BASE_MODEL_CLASS_NAME = 'Organisation'  # A string to prevent circular imports.

    # Base path.
    _PATH_BASE = '/organisations/{organisation_id}/process_executions/{process_execution_id}'

    # Override the standard model paths.
    _PATH_DELETE = _PATH_BASE
    _PATH_GET = _PATH_BASE
    _PATH_PATCH = _PATH_BASE

    # Add in the custom model paths.
    _PATH_COMPONENTS = f"{_PATH_BASE}/process_service_executions"

    def check_complete(self, wait=False):
        """
        Checks whether the execution has completed. Optionally waits for the execution to complete.

        Args:
            wait: Optionally wait for the execution to complete? Default False.

        Returns:
            True if the execution is complete.

        Raises:
            RequestError: if any request fails.
            ModelError: if the execution failed.
            ModelError: if a model could not be loaded or validated from the Fusion Platform<sup>&reg;</sup>.
        """
        complete = False

        # Optionally wait for the execution to finish.
        while not complete:
            # Load in the most recent version of the model.
            self.get(organisation_id=self.organisation_id)

            # See if the execution has completed.
            self._logger.debug('checking for execution %s to complete: %f', self.id, self.progress)
            complete = self.progress >= 100

            # Raise an exception if the execution failed.
            if complete and not self.success:
                raise ModelError(i18n.t('models.process_execution.execution_failed'))

            if complete:
                self._logger.debug('execution %s is complete', self.id)

            if (not wait) or complete:
                break

            # We are waiting, so block for a short while.
            sleep(self.__class__._API_UPDATE_WAIT_PERIOD)

        return complete

    @property
    def components(self):
        """
        Provides an iterator through the process execution's components.

        Returns:
            An iterator through the component objects.

        Raises:
            RequestError: if any get fails.
            ModelError: if a model could not be loaded or validated from the Fusion Platform<sup>&reg;</sup>.
        """
        return ProcessServiceExecution._models_from_api_path(self._session, self._get_path(self.__class__._PATH_COMPONENTS))



"""
Main entry point for package.

author: Matthew Casey

&copy; [Digital Content Analysis Technology Ltd](https://www.d-cat.co.uk)
"""

from fusion_platform.base import Base

# Main entry point when the package is executed.
if __name__ == "__main__":
    Base.main()



"""
Process service execution log model class file.

author: Matthew Casey

&copy; [Digital Content Analysis Technology Ltd](https://www.d-cat.co.uk)
"""

from marshmallow import Schema, EXCLUDE

from fusion_platform.models import fields
from fusion_platform.models.model import Model


# Define the model schema classes. These are maintained from the API definitions.

class ProcessServiceExecutionLogSchema(Schema):
    """
    Schema class for process service execution log model.

    Each process service execution log model has the following fields (and nested fields):

    .. include::process_service_execution_log.md
    """
    id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    created_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    updated_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    process_service_execution_id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    logged_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    message = fields.String(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ProcessServiceExecutionLog(Model):
    """
    Process service execution log model class providing attributes and methods to manipulate process execution item details.
    """

    # Override the schema.
    _SCHEMA = ProcessServiceExecutionLogSchema()



"""
General utilities.

author: Matthew Casey

&copy; [Digital Content Analysis Technology Ltd](https://www.d-cat.co.uk)
"""

from datetime import date, datetime, time, timezone
from decimal import Decimal
from types import MappingProxyType


def dict_nested_get(dictionary_or_value, keys, default=None):
    """
    Performs a dictionary.get(key, default) using the supplied list of keys assuming that each successive key is nested.

    For example, for a dictionary dictionary = { "key1": { "key2": 1 } }, use nested_get(dictionary, ["key1", "key2"]) to get the value of "key2".

    Args:
        dictionary_or_value: The dictionary to get the value from or the value itself from this recursive method.
        keys: The list of nested keys.
        default: The default value to return if no value exists. Default is None.

    Returns:
        The value of the nested key or the default if not found.
    """
    if isinstance(dictionary_or_value, dict) and isinstance(keys, list) and (len(keys) > 1):
        key = keys.pop(0)
        return dict_nested_get(dictionary_or_value.get(key, default), keys, default)
    elif isinstance(dictionary_or_value, dict) and isinstance(keys, list) and (len(keys) == 1):
        return dictionary_or_value.get(keys[0], default)
    elif (dictionary_or_value is not None) and (not isinstance(dictionary_or_value, dict)):
        return dictionary_or_value
    else:
        return default


def json_default(value):
    """
    Used during JSON serialisation for objects which cannot be serialised. This will attempt to convert the values and otherwise return a string.

    Args:
        value: The value to serialise.

    Returns:
        The serialised value.
    """
    if isinstance(value, time) or isinstance(value, date) or isinstance(value, datetime):
        # Convert datetimes. Note that if the datetime does not have a timezone, we enforce UTC. See also fields.DateTime#_serialize.
        value = value.astimezone(timezone.utc) if isinstance(value, datetime) and (value.tzinfo is None) else value
        return value.isoformat()
    elif isinstance(value, Decimal):
        # We want to output numbers as either integers (no trailing ".0") or floats. This will also convert decimals.
        return int(value) if float(value).is_integer() else float(value)
    elif isinstance(value, MappingProxyType):
        # Read-only dictionaries may be defined using the MappingProxyType. For serialisation, we therefore convert them back to dictionaries.
        return dict(value)
    else:
        return str(value)


def string_camel_to_underscore(string):
    """
    Converts a camel case string to underscore case.

    Args:
        string: The string to convert.

    Returns:
        The converted string.
    """

    return ''.join(['_' + i.lower() if i.isupper() else i for i in string]).lstrip('_').lower() if string is not None else None


def value_from_read_only(value):
    """
    Takes a value and makes it writable. This recursive method will deal with dictionaries and lists. Here, dictionaries are assumed to be mapping proxies, while
    tuples are assumed to be lists.

    Args:
        value: The value to make read-only.

    Returns:
        The read-only value.
    """
    if isinstance(value, MappingProxyType):
        return dict({inner_key: value_from_read_only(inner_value) for inner_key, inner_value in value.items()})  # Creates a mutable mapping.
    elif isinstance(value, tuple):  # This assumes that tuples are really lists.
        return list([value_from_read_only(inner_value) for inner_value in value])  # Creates a mutable list.
    else:
        return value


def value_to_read_only(value):
    """
    Takes a value and makes it read-only. This recursive method will deal with dictionaries and lists. This method will not deal with objects that are immutable,
    such as datetimes, but rather makes sure that the references to objects cannot be changed. For example, dictionary values cannot be replaced or list items
    removed.

    Args:
        value: The value to make read-only.

    Returns:
        The read-only value.
    """
    if isinstance(value, dict):
        return MappingProxyType({inner_key: value_to_read_only(inner_value) for inner_key, inner_value in value.items()})  # Creates an immutable mapping.
    elif isinstance(value, list):
        return tuple([value_to_read_only(inner_value) for inner_value in value])  # Creates an immutable list, which is just a tuple.
    else:
        return value


def value_to_string(value):
    """
    Used to convert any value into a useful string representation. For example, datetimes into ISO format. Also handles other type conversions to make them pretty.

    Args:
        value: The value to convert to a string.

    Returns:
        The corresponding (pretty) string value.
    """
    if isinstance(value, list) or isinstance(value, tuple):
        return f"[{', '.join([value_to_string(inner) for inner in value])}]"
    elif isinstance(value, dict):
        items = [f"{key}: {value_to_string(inner)}" for key, inner in value.items()]
        return f"{{{', '.join(items)}}}"
    else:
        # Return whatever is appropriate for JSON, but as a string.
        return str(json_default(value))



"""
Compiled translations.

author: Matthew Casey

&copy; [Digital Content Analysis Technology Ltd](https://www.d-cat.co.uk)
"""

# Do not modify this file manually as it is built automatically by the localisations.py script.

import i18n

# @formatter:off
i18n.add_translation('session.request_failed', 'API request failed: %{message}', 'en')
i18n.add_translation('session.login_failed', 'Login failed', 'en')
i18n.add_translation('session.missing_password', 'Password must be specified', 'en')
i18n.add_translation('session.missing_email_user_id', 'Either an email address or a user id must be specified', 'en')
i18n.add_translation('fusion_platform.support', 'Support: support@d-cat.co.uk', 'en')
i18n.add_translation('fusion_platform.version_date', 'Date: %{version_date}', 'en')
i18n.add_translation('fusion_platform.version', 'Version: %{version}', 'en')
i18n.add_translation('fusion_platform.sdk', 'Fusion Platform(r) SDK', 'en')
i18n.add_translation('models.data_file.failed_download_url', 'Failed to get URL from download file response', 'en')
i18n.add_translation('models.data_file.no_download', 'No download is in progress', 'en')
i18n.add_translation('models.data_file.download_already_in_progress', 'Cannot download file as the download is already in progress', 'en')
i18n.add_translation('models.data_file.organisation_id.description', 'The owning organisation.', 'en')
i18n.add_translation('models.data_file.organisation_id.title', 'Organisation', 'en')
i18n.add_translation('models.data.no_create', 'No create is in progress', 'en')
i18n.add_translation('models.data.failed_add_missing_file', 'Failed to add file as the file does not exist: %{file}', 'en')
i18n.add_translation('models.data.failed_add_file_not_unique', 'Failed to add file as the id is not unique', 'en')
i18n.add_translation('models.data.failed_add_file_url', 'Failed to get URL from add file response', 'en')
i18n.add_translation('models.data.failed_add_file_id', 'Failed to get id from add file response', 'en')
i18n.add_translation('models.data.failed_copy_id', 'Failed to get data id from copy response', 'en')
i18n.add_translation('models.process_execution.execution_failed', 'Execution has failed', 'en')
i18n.add_translation('models.fields.uuid.invalid_uuid', 'Not a valid utf-8 string', 'en')
i18n.add_translation('models.fields.url.invalid_url', 'Not a valid URL', 'en')
i18n.add_translation('models.fields.tuple.invalid', 'Not a valid tuple', 'en')
i18n.add_translation('models.fields.timedelta.invalid', 'Not a valid period of time', 'en')
i18n.add_translation('models.fields.string.invalid_utf8', 'Not a valid utf-8 string', 'en')
i18n.add_translation('models.fields.string.invalid', 'Not a valid string', 'en')
i18n.add_translation('models.fields.relativedelta.invalid', 'Not a valid relative period of time', 'en')
i18n.add_translation('models.fields.nested.type', 'Invalid type', 'en')
i18n.add_translation('models.fields.list.invalid', 'Not a valid list', 'en')
i18n.add_translation('models.fields.ip.invalid_ip', 'Not a valid IP address', 'en')
i18n.add_translation('models.fields.integer.too_large', 'Integer too large', 'en')
i18n.add_translation('models.fields.integer.invalid', 'Not a valid integer', 'en')
i18n.add_translation('models.fields.float.special', 'Special numeric values (nan or infinity) are not permitted.', 'en')
i18n.add_translation('models.fields.float.too_large', 'Float too large', 'en')
i18n.add_translation('models.fields.float.invalid', 'Not a valid float', 'en')
i18n.add_translation('models.fields.email.invalid', 'Not a valid email address', 'en')
i18n.add_translation('models.fields.dict.invalid', 'Not a valid dictionary', 'en')
i18n.add_translation('models.fields.decimal.special', 'Special numeric values (nan or infinity) are not permitted', 'en')
i18n.add_translation('models.fields.decimal.too_large', 'Decimal too large', 'en')
i18n.add_translation('models.fields.decimal.invalid', 'Not a valid decimal', 'en')
i18n.add_translation('models.fields.datetime.format', '\'{input}\' cannot be formatted as a {obj_type}', 'en')
i18n.add_translation('models.fields.datetime.invalid_awareness', 'Not a valid {awareness} {obj_type}', 'en')
i18n.add_translation('models.fields.datetime.invalid', 'Not a valid {obj_type}', 'en')
i18n.add_translation('models.fields.boolean.invalid', 'Not a valid boolean', 'en')
i18n.add_translation('models.model.update_empty_body', 'Update cannot be requested as there are no attributes to be used (read-only attributes have been removed)', 'en')
i18n.add_translation('models.model.create_empty_body', 'Create cannot be requested as there are no attributes to be used (read-only attributes have been removed)', 'en')
i18n.add_translation('models.model.failed_model_validation', 'Failed to validate model: %{message}', 'en')
i18n.add_translation('models.model.failed_model_new', 'Failed to get model template from response', 'en')
i18n.add_translation('models.model.failed_model_send_and_load', 'Failed to request and load model', 'en')
i18n.add_translation('models.model.no_such_keys', 'No such keys %{keys}', 'en')
i18n.add_translation('models.model.readonly_property', 'Property %{property} is read-only and cannot be set', 'en')
i18n.add_translation('models.model.not_persisted', 'Model is not persisted in the Fusion Platform(r)', 'en')
i18n.add_translation('models.model.already_persisted', 'Model is already persisted in the Fusion Platform(r)', 'en')
i18n.add_translation('models.process.failed_copy', 'Failed to get process from copy response', 'en')
i18n.add_translation('models.process.execution_should_have_started', 'Process execution should have started by now', 'en')
i18n.add_translation('models.process.wrong_file_type', 'File type of supplied data object (%{actual}) does not match the file type for the input (%{expected})', 'en')
i18n.add_translation('models.process.data_not_ready', 'Data object is not ready to be used in a process', 'en')
i18n.add_translation('models.process.option_wrong_type', 'Option value should be of type %{type}', 'en')
i18n.add_translation('models.process.cannot_find_option', 'No such option', 'en')
i18n.add_translation('models.process.cannot_find_input', 'No such input', 'en')
i18n.add_translation('models.process.option_not_specified', 'Option name or object must be provided to set option', 'en')
i18n.add_translation('models.process.data_not_specified', 'Data object must be provided to set input', 'en')
i18n.add_translation('models.process.input_not_specified', 'Input number or object must be provided to set input', 'en')
i18n.add_translation('models.process.no_change_executing', 'Process cannot be modified as it is currently executing', 'en')
i18n.add_translation('models.process.option.constrained_values.description', 'The constrained values for the option.', 'en')
i18n.add_translation('models.process.option.constrained_values.title', 'Constrained Values', 'en')
i18n.add_translation('models.process.option.constrained_names.description', 'The constrained value names for the option.', 'en')
i18n.add_translation('models.process.option.constrained_names.title', 'Constrained Names', 'en')
# @formatter:on



"""
Process model class file.

author: Matthew Casey

&copy; [Digital Content Analysis Technology Ltd](https://www.d-cat.co.uk)
"""

from datetime import datetime, timedelta, timezone
from functools import partial
import i18n
from marshmallow import Schema, EXCLUDE
from time import sleep

import fusion_platform
from fusion_platform.common.utilities import value_from_read_only
from fusion_platform.models import fields
from fusion_platform.models.data import Data
from fusion_platform.models.model import Model, ModelError
from fusion_platform.models.process_execution import ProcessExecution
from fusion_platform.session import Session


# Define a schema class used to coerce option values. See #__coerce_value.

class OptionDataTypeSchema(Schema):
    """
    Defines a Marshmallow schema which can be used to coerce any option value into its underlying Python data type.
    """

    # List out all the allowed field types using their corresponding data type names.
    numeric = fields.Float(allow_none=True)
    currency = fields.Decimal(allow_none=True)
    boolean = fields.Boolean(allow_none=True)
    datetime = fields.DateTime(allow_none=True)
    string = fields.String(allow_none=True)
    constrained = fields.String(allow_none=True)


# Define a model class for an option which overrides the representation to display more relevant information.

class Option(Model):

    def __repr__(self):
        """
        Returns:
            A string representation of the object.
        """
        return f"{self.title} ('{self.name}', {self.data_type}{', required' if self.required else ''}) = {self.value}: {self.description}"


# Define a model class for an input which overrides the representation to display more relevant information.

class Input(Model):

    def __repr__(self):
        """
        Returns:
            A string representation of the object.
        """
        return f"{self.title} ({self.file_type}) = {self.name + ' (' + str(self.id) + ')' if hasattr(self, self.__class__._FIELD_NAME) and hasattr(self, self.__class__._FIELD_ID) else None}: {self.description}"


# Define the model schema classes. These are maintained from the API definitions.

class ProcessChainOptionSchema(Schema):
    """
    Nested schema class for SSD chain option.
    """
    name = fields.String(required=True)
    value = fields.String(allow_none=True)
    data_type = fields.String(required=True)
    validation = fields.String(allow_none=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ProcessChainSchema(Schema):
    """
    Nested schema class for SSD chain.
    """
    ssd_id = fields.UUID(required=True)
    service_id = fields.UUID(required=True)
    inputs = fields.List(fields.UUID(allow_none=True), allow_none=True)
    outputs = fields.List(fields.UUID(required=True), allow_none=True)
    options = fields.List(fields.Nested(ProcessChainOptionSchema()), allow_none=True)
    intermediate = fields.Boolean(allow_none=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ProcessExecutionStatusSchema(Schema):
    """
    Nested schema class for execution status.
    """
    process_execution_id = fields.UUID(required=True)
    job_status = fields.String(required=True)
    started_at = fields.DateTime(required=True)
    ended_at = fields.DateTime(allow_none=True)
    delete_warning_status = fields.String(required=True)
    delete_expiry = fields.DateTime(required=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ProcessSelectorSchema(Schema):
    """
    Nested schema class for selector, category, data type, unit and format.
    """
    selector = fields.String(required=True)
    category = fields.String(required=True)
    data_type = fields.String(required=True)
    unit = fields.String(allow_none=True)
    validation = fields.String(allow_none=True)

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ProcessInputSchema(Schema):
    """
    Nested schema class for the processes inputs.
    """
    ssd_id = fields.UUID(required=True)
    input = fields.Integer(required=True)
    file_type = fields.String(required=True)
    resolution = fields.Integer(allow_none=True)
    selectors = fields.List(fields.Nested(ProcessSelectorSchema()), allow_none=True)
    id = fields.UUID(required=True)
    model = fields.String(required=True)
    change_trigger = fields.Boolean(required=True)
    change_hash = fields.String(allow_none=True)
    title = fields.String(allow_none=True)
    description = fields.String(allow_none=True)

    name = fields.String(allow_none=True)  # Changed to add this field for displaying the input name.

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ProcessOptionSchema(Schema):
    """
    Nested schema class for options which are provided to the SSD images when run.
    """
    ssd_id = fields.UUID(required=True)
    name = fields.String(required=True)
    value = fields.String(allow_none=True)
    required = fields.Boolean(required=True)
    data_type = fields.String(required=True)
    validation = fields.String(allow_none=True)
    mutually_exclusive = fields.String(allow_none=True)
    advanced = fields.Boolean(allow_none=True)
    title = fields.String(allow_none=True)
    description = fields.String(allow_none=True)

    constrained_names = fields.List(fields.String(required=True), allow_none=True, metadata={'title': i18n.t('models.process.option.constrained_names.title'),
                                                                                             'description': i18n.t(
                                                                                                 'models.process.option.constrained_names.description')})  # Added this field to gold extracted data.
    constrained_values = fields.List(fields.String(required=True), allow_none=True, metadata={'title': i18n.t('models.process.option.constrained_values.title'),
                                                                                              'description': i18n.t(
                                                                                                  'models.process.option.constrained_values.description')})  # Added this field to gold extracted data.

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class ProcessSchema(Schema):
    """
    Schema class for process model.

    Each process model has the following fields (and nested fields):

    .. include::process.md
    """
    id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    created_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    updated_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    organisation_id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    ssd_id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    service_id = fields.UUID(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    name = fields.String(required=True)

    inputs = fields.List(fields.Nested(ProcessInputSchema()), allow_none=True, metadata={'hide': True})  # Changed to hide as an attribute.
    options = fields.List(fields.Nested(ProcessOptionSchema()), allow_none=True, metadata={'hide': True})  # Changed to hide as an attribute.
    chains = fields.List(fields.Nested(ProcessChainSchema()), allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    # Removed maximum_bounds.

    run_type = fields.String(required=True)
    repeat_count = fields.Integer(required=True)
    repeat_start = fields.DateTime(required=True)
    repeat_end = fields.DateTime(allow_none=True)
    repeat_gap = fields.RelativeDelta(allow_none=True)
    repeat_offset = fields.TimeDelta(allow_none=True)

    process_status = fields.String(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    process_status_at = fields.DateTime(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    process_status_changed_by = fields.UUID(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    output_storage_period = fields.Integer(allow_none=True)
    test_run = fields.Boolean(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    # Removed prices.
    price = fields.Decimal(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.

    deletable = fields.String(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
    has_executions = fields.String(required=True, metadata={'hide': True})  # Changed to hide as an attribute.
    has_executions_organisation_id = fields.String(required=True, metadata={'hide': True})  # Changed to hide as an attribute.
    executions = fields.List(fields.Nested(ProcessExecutionStatusSchema()), allow_none=True, metadata={'hide': True})  # Changed to hide as an attribute.

    # Removed creator.
    # Removed search.

    class Meta:
        """
        When loading an object, make sure we exclude any unknown fields, rather than raising an exception, and put fields in their definition order.
        """
        unknown = EXCLUDE
        ordered = True


class Process(Model):
    """
    Process model class providing attributes and methods to manipulate process item details.
    """

    # Override the schema.
    _SCHEMA = ProcessSchema()

    # Override the base model class name.
    _BASE_MODEL_CLASS_NAME = 'Organisation'  # A string to prevent circular imports.

    # Base path.
    _PATH_ROOT = '/organisations/{organisation_id}/processes'
    _PATH_BASE = f"{_PATH_ROOT}/{{process_id}}"

    # Override the standard model paths.
    _PATH_COPY = f"{_PATH_BASE}/copy"
    _PATH_CREATE = _PATH_ROOT
    _PATH_DELETE = _PATH_BASE
    _PATH_GET = _PATH_BASE
    _PATH_NEW = f"{_PATH_ROOT}/new"
    _PATH_PATCH = _PATH_BASE

    # Add in the custom model paths.
    _PATH_EXECUTE = f"{_PATH_BASE}/execute"
    _PATH_EXECUTIONS = f"{_PATH_BASE}/executions"
    _PATH_STOP = f"{_PATH_BASE}/stop"

    # Process status values.
    _PROCESS_STATUS_EXECUTE = 'execute'

    # The maximum number of seconds to wait after an execution was meant to start.
    _EXECUTE_WAIT_TOLERANCE = 300

    # Validation parsing for constrained values.
    _VALIDATION_DELIMITER = ';'
    _VALIDATION_EQUALS = '='
    _VALIDATION_FORMAT = 'format'
    _VALIDATION_ITEM_DELIMITER = ','
    _VALIDATION_NAMES = 'names'
    _VALIDATION_VALUES = 'values'

    # Allowed file type substitutions.
    # @formatter:off
    _FILE_TYPE_SUBSTITUTIONS = {
        fusion_platform.FILE_TYPE_GEOTIFF: [fusion_platform.FILE_TYPE_GEOTIFF, fusion_platform.FILE_TYPE_DEM, fusion_platform.FILE_TYPE_JPEG2000],
        fusion_platform.FILE_TYPE_JPEG2000: [fusion_platform.FILE_TYPE_JPEG2000],
        fusion_platform.FILE_TYPE_DEM: [fusion_platform.FILE_TYPE_DEM],
        fusion_platform.FILE_TYPE_GEOJSON: [fusion_platform.FILE_TYPE_GEOJSON, fusion_platform.FILE_TYPE_KML, fusion_platform.FILE_TYPE_KMZ,
                                            fusion_platform.FILE_TYPE_ESRI_SHAPEFILE],
        fusion_platform.FILE_TYPE_KML: [fusion_platform.FILE_TYPE_KML],
        fusion_platform.FILE_TYPE_KMZ: [fusion_platform.FILE_TYPE_KMZ],
        fusion_platform.FILE_TYPE_CSV: [fusion_platform.FILE_TYPE_CSV],
        fusion_platform.FILE_TYPE_ESRI_SHAPEFILE: [fusion_platform.FILE_TYPE_ESRI_SHAPEFILE],
        fusion_platform.FILE_TYPE_JPEG: [fusion_platform.FILE_TYPE_JPEG],
        fusion_platform.FILE_TYPE_PNG: [fusion_platform.FILE_TYPE_PNG],
        fusion_platform.FILE_TYPE_OTHER: [fusion_platform.FILE_TYPE_OTHER],
    }
    # @formatter:on

    @classmethod
    def __coerce_value(cls, value, data_type):
        """
        Attempts to coerce a value into its corresponding data type.

        Args:
            value: The value to coerce.
            data_type: The required data type.

        Returns:
            The coerced value of the correct data type.
        """
        # Deal with None values.
        if (value is None) or (value == str(None)) or (value == 'null'):
            return None
        else:
            # Use the option schema to attempt to load the value using its data type as a name.
            model = OptionDataTypeSchema().load({data_type: value})
            return model.get(data_type)

    def copy(self, name):
        """
        Creates a new template process from an existing process object. This process is not persisted to the Fusion Platform<sup>&reg;</sup>.

        Args:
            name: The name of the copy.

        Returns:
            The new template process object.

        Raises:
            RequestError: if the copy fails.
            ModelError: if the model could not be created and validated by the Fusion Platform<sup>&reg;</sup>.
        """
        # Get a new template for the process model using the process as the source of the copy.
        process = Process(self._session)

        # Copy the process model.
        response = self._session.request(path=self._get_path(self.__class__._PATH_COPY), method=Session.METHOD_GET)

        # Assume that the resulting model is held within the expected key within the resulting dictionary.
        if Model._RESPONSE_KEY_MODEL not in response:
            raise ModelError(i18n.t('models.process.failed_copy'))

        # Make sure we have an organisation id.
        modified_response = response.get(Model._RESPONSE_KEY_MODEL, {})
        modified_response['organisation_id'] = self.organisation_id

        # Load the response into the model. We ignore missing required fields and those which are None.
        process._set_model_from_response(modified_response, partial=True)

        # Update the options because these are not copied across to the new model, even though they are returned in the response.
        options = modified_response.get('options', [])
        options = [] if options is None else options

        for option in options:
            # Set the option value. The value may be a string, rather than the required type, but that will automatically be converted.
            if (option.get(self.__class__._FIELD_NAME) is not None) and (option.get(self.__class__._FIELD_VALUE) is not None):
                process.__set_option(name=option.get(self.__class__._FIELD_NAME), value=option.get(self.__class__._FIELD_VALUE), coerce_value=True)

        # And now the same for the inputs. Here we also validate each input by retrieving it.
        inputs = modified_response.get('inputs', [])
        inputs = [] if inputs is None else inputs

        for i, input in enumerate(inputs):
            # See if we can find the associated model.
            if input.get(self.__class__._FIELD_ID) is not None:
                data = Data._model_from_api_id(self._session, organisation_id=self.organisation_id, data_id=input.get(self.__class__._FIELD_ID))
                process.__set_input(number=i + 1, data=data)

        # Set the name.
        process.update(name=name)

        # Return the copy.
        return process

    def create(self):
        """
        Attempts to persist the template process in the Fusion Platform<sup>&reg;</sup> so that it can be executed.

        Raises:
            RequestError: if the create fails.
            ModelError: if the model could not be created and validated by the Fusion Platform<sup>&reg;</sup>.
        """
        # Attempt to issue the create.
        self._create()

    def execute(self, wait=False):
        """
        Attempts to execute the created process in the Fusion Platform<sup>&reg;</sup>. Optionally waits for the next execution to start, and then for it to complete.

        Args:
            wait: Optionally wait for the next execution to start and complete? Default False.

        Raises:
            RequestError: if the execute fails.
            ModelError: if the execution fails.
        """
        # Send the request and load the resulting model.
        self._send_and_load(self._get_path(self.__class__._PATH_EXECUTE), method=Session.METHOD_POST)

        if wait:
            # If we are waiting for the execution to complete, wait for the next execution to start...
            self.wait_for_next_execution()

            # ...and for all the executions to complete.
            for execution in self.executions:
                execution.check_complete(wait=wait)

    @property
    def executions(self):
        """
        Provides an iterator through the process's executions.

        Returns:
            An iterator through the execution objects.
        
        Raises:
            RequestError: if any get fails.
            ModelError: if a model could not be loaded or validated from the Fusion Platform<sup>&reg;</sup>.
        """
        return ProcessExecution._models_from_api_path(self._session, self._get_path(self.__class__._PATH_EXECUTIONS), reverse=True)  # Most recent first.

    @classmethod
    def __extract_constrained_validation(cls, validation):
        """
        Extracts the constrained values from a constrained option validation.

        Args:
            validation: The constrained option validation.

        Returns:
            A tuple (constrained_names, constrained_values) of the extracted elements or (None, None) if the constrained values cannot be extracted.
        """
        # Extract the elements.
        elements = validation.split(Process._VALIDATION_DELIMITER)
        constrained_names = None
        constrained_values = None

        if len(elements) <= 2:
            for element in elements:
                values = element.split(Process._VALIDATION_EQUALS)

                if len(values) > 0:
                    lhs = values[0].lower()
                    rhs = Process._VALIDATION_EQUALS.join(values[1:])

                    try:
                        if lhs == Process._VALIDATION_VALUES:
                            constrained_values = rhs.split(Process._VALIDATION_ITEM_DELIMITER)

                        if lhs == Process._VALIDATION_NAMES:
                            constrained_names = rhs.split(Process._VALIDATION_ITEM_DELIMITER)
                    except:
                        pass  # Cannot be parsed.

        return constrained_names, constrained_values

    @classmethod
    def __extract_datetime_validation(cls, validation):
        """
        Extracts the datetime format from a datatime option validation.

        Args:
            validation: The datatime option validation.

        Returns:
            The datetime format, or None if it cannot be extracted.
        """
        # Extract the elements.
        elements = validation.split(Process._VALIDATION_DELIMITER)
        format = None

        if len(elements) <= 3:
            for element in elements:
                values = element.split(Process._VALIDATION_EQUALS)

                if len(values) > 0:
                    lhs = values[0].lower()
                    rhs = Process._VALIDATION_EQUALS.join(values[1:])

                    try:
                        if lhs == Process._VALIDATION_FORMAT:
                            format = rhs
                    except:
                        pass  # Cannot be parsed.

        return format

    def find_executions(self, id=None, group_id=None):
        """
        Searches for the process's executions with the specified id and/or group id, returning the first object found and an iterator.

        Args:
            id: The execution id to search for.
            group_id: The execution group id to search for.

        Returns:
            The first found execution object, or None if not found, and an iterator through the found execution objects.
        
        Raises:
            RequestError if any get fails.
            ModelError if a model could not be loaded or validated from the Fusion Platform<sup>&reg;</sup>.
        """
        filter = self.__class__._build_filter(
            [(self.__class__._FIELD_ID, self.__class__._FILTER_MODIFIER_EQ, id), (self.__class__._FIELD_GROUP_ID, self.__class__._FILTER_MODIFIER_EQ, group_id)])

        # Build the partial find generator and execute it.
        find = partial(ProcessExecution._models_from_api_path, self._session, self._get_path(self.__class__._PATH_EXECUTIONS), filter=filter)
        return self.__class__._first_and_generator(find)

    @property
    def inputs(self):
        """
        Provides an iterator through the process' inputs.

        Returns:
            An iterator through the inputs.
        """
        for input in self._model.get(self.__class__._FIELD_INPUTS, []):
            # We first have to remove the mapping proxy so that we can wrap the dictionary in a model.
            input = value_from_read_only(input)

            # See if we can find the associated model.
            if input.get(self.__class__._FIELD_ID) is not None:
                data = Data._model_from_api_id(self._session, organisation_id=self.organisation_id, id=input.get(self.__class__._FIELD_ID))
                input[self.__class__._FIELD_NAME] = data.name

            # Encapsulate the dictionary within a model (which does not talk to the API).
            model = Input(None, schema=ProcessInputSchema())
            model._set_model(input)

            yield model

    @property
    def options(self):
        """
        Provides an iterator through the process' options.

        Returns:
            An iterator through the options.
        """
        for option in self._model.get(self.__class__._FIELD_OPTIONS, []):
            # We first have to remove the mapping proxy so that we can wrap the dictionary in a model.
            option = value_from_read_only(option)

            # If the option is a constrained data type, then add in the constrained names and values from the validation.
            if option.get(self.__class__._FIELD_DATA_TYPE) == fusion_platform.DATA_TYPE_CONSTRAINED:
                option[self.__class__._FIELD_CONSTRAINED_NAMES], option[self.__class__._FIELD_CONSTRAINED_VALUES] = self.__class__.__extract_constrained_validation(
                    option.get(self.__class__._FIELD_VALIDATION, ''))

            # Coerce the value to be of the correct data type.
            option[self.__class__._FIELD_VALUE] = self.__class__.__coerce_value(option.get(self.__class__._FIELD_VALUE),
                                                                                option.get(self.__class__._FIELD_DATA_TYPE))

            # Encapsulate the dictionary within a model (which does not talk to the API).
            model = Option(None, schema=ProcessOptionSchema())
            model._set_model(option)

            yield model

    def __set_input(self, number=None, input=None, data=None):
        """
        Sets the specified input for the process to the data item. An exception is raised if the process is in the execute status, the input does not exist, is not
        ready to be used or has the wrong file type.

        Args:
            number: The input number to set, starting from 1 for the first input. Either the number or the input must be provided.
            input: The input object for the input to set. Either the number or the input must be provided.
            data: The data object to use for the input.

        Raises:
            ModelError: if the process is the execute status.
            ModelError: if the input does not exist.
            ModelError: if the data object is ready to be used in a process.
            ModelError: if the data object has a different file type to the input.
        """
        # Make sure the arguments are provided.
        if (number is None) and (input is None):
            raise ModelError(i18n.t('models.process.input_not_specified'))

        if (data is None) or (not isinstance(data, Data)):
            raise ModelError(i18n.t('models.process.data_not_specified'))

        # Make sure the process is not in the execute state.
        if hasattr(self, self.__class__._FIELD_PROCESS_STATUS) and (self.process_status == Process._PROCESS_STATUS_EXECUTE):
            raise ModelError(i18n.t('models.process.no_change_executing'))

        # Find the corresponding input.
        index = None
        found_input = None

        for index, item in enumerate(self._model.get(self.__class__._FIELD_INPUTS, [])):
            found = (number is not None) and (number == (index + 1))
            found = (input is not None) and (str(input.ssd_id) == str(item.get(self.__class__._FIELD_SSD_ID))) and (
                    input.input == item.get(self.__class__._FIELD_INPUT)) if not found else found

            if found:
                found_input = item
                break

        if found_input is None:
            raise ModelError(i18n.t('models.process.cannot_find_input'))

        # Check that all the files in the data object are ready to be used. Along the way, pick out the first file type.
        found_file_type = None
        ready = True

        for file in data.files:
            found_file_type = file.file_type if found_file_type is None else found_file_type

            if not hasattr(file, self.__class__._FIELD_PUBLISHABLE):
                ready = False

        if not ready:
            raise ModelError(i18n.t('models.process.data_not_ready'))

        # Check the file type against the allowed list of substitutions.
        if found_file_type not in Process._FILE_TYPE_SUBSTITUTIONS.get(found_input.get(self.__class__._FIELD_FILE_TYPE), []):
            raise ModelError(i18n.t('models.process.wrong_file_type', expected=found_input.get(self.__class__._FIELD_FILE_TYPE), actual=found_file_type))

        # We can now update the input.
        self._set_field([self.__class__._FIELD_INPUTS, index, self.__class__._FIELD_ID], data.id)

    def __set_option(self, name=None, option=None, value=None, coerce_value=False):
        """
        Sets the specified option for the process to the value. An exception is raised if the process is in the execute status, the option does not exist or the
        value has the wrong type.

        Args:
            name: The option name to set. Either the name or the option must be provided.
            option: The option object for the option to set. Either the name or the option must be provided.
            value: The value for the option.
            coerce_value: Optionally coerce the supplied value to be the correct type. Default False.

        Raises:
            ModelError: if the process is the execute status.
            ModelError: if the value has a different type to the option.
        """
        # Make sure the arguments are provided.
        if (name is None) and (option is None):
            raise ModelError(i18n.t('models.process.option_not_specified'))

        # Make sure the process is not in the execute state.
        if hasattr(self, self.__class__._FIELD_PROCESS_STATUS) and (self.process_status == Process._PROCESS_STATUS_EXECUTE):
            raise ModelError(i18n.t('models.process.no_change_executing'))

        # Find the corresponding option.
        index = None
        found_option = None

        for index, item in enumerate(self._model.get(self.__class__._FIELD_OPTIONS, [])):
            found = (name is not None) and (name == item.get(self.__class__._FIELD_NAME))
            found = (option is not None) and (str(option.ssd_id) == str(item.get(self.__class__._FIELD_SSD_ID))) and (
                    option.name == item.get(self.__class__._FIELD_NAME)) if not found else found

            if found:
                found_option = item
                break

        if found_option is None:
            raise ModelError(i18n.t('models.process.cannot_find_option'))

        # Check that the option has the same data type as the value. We cannot check this if either value is None. Note that supplied ints can be used for floats.
        data_type = found_option.get(self.__class__._FIELD_DATA_TYPE)
        existing_value = self.__class__.__coerce_value(found_option[self.__class__._FIELD_VALUE], data_type)

        # Optionally coerce the supplied value, now that we know its data type.
        if coerce_value:
            value = self.__class__.__coerce_value(value, data_type)

        class_matches = isinstance(value, existing_value.__class__)

        if isinstance(value, int) and isinstance(existing_value, float):
            class_matches = True

        if (value is not None) and (existing_value is not None) and (not class_matches):
            raise ModelError(i18n.t('models.process.option_wrong_type', type=existing_value.__class__))

        # We can now update the option. All options are expressed as strings with the correct format.
        validation = found_option.get(self.__class__._FIELD_VALIDATION)
        self._set_field([self.__class__._FIELD_OPTIONS, index, self.__class__._FIELD_VALUE], self.__class__.__value_to_option(value, data_type, validation))

    def stop(self):
        """
        Stops the process from being executed. This will abort any executions which are in progress and prevent any scheduled executions from taking place.

        Raises:
            RequestError: if the stop fails.
        """
        # Send the request and load the resulting model.
        self._send_and_load(self._get_path(self.__class__._PATH_STOP), method=Session.METHOD_POST)

    def update(self, input_number=None, input=None, data=None, option_name=None, option=None, value=None, coerce_value=False, **kwargs):
        """
        Attempts to update the model object with the given values. This assumes the model is updated using a PATCH RESTful request. This assumes that the patch body
        contains key names which include the name of the model class. Overridden to prevent changes if the process is being executed and to handle the special cases
        of setting inputs and options.

        Args:
            input_number: The input number to set, starting from 1 for the first input. Either the number or the input must be provided when setting an input.
            input: The input object for the input to set. Either the number or the input must be provided when setting an input.
            data: The data object to use for an input.
            option_name: The option name to set. Either the name or the option must be provided when setting an option.
            option: The option object for the option to set. Either the name or the option must be provided when setting an option.
            value: The value for the option.
            coerce_value: Optionally coerce the supplied value to be the correct type. Default False.
            kwargs: The model attributes which are to be patched.

        Raises:
            RequestError: if the update fails.
            ModelError: if the process is in the execute state.
            ModelError: if the process has been persisted and changes are requested to an input or option.
            ModelError: if the model could not be loaded or validated from the Fusion Platform<sup>&reg;</sup>.
        """
        # Make sure the process is not in the execute state.
        if hasattr(self, self.__class__._FIELD_PROCESS_STATUS) and (self.process_status == Process._PROCESS_STATUS_EXECUTE):
            raise ModelError(i18n.t('models.process.no_change_executing'))

        # Deal with the special case of inputs.
        if (input_number is not None) or (input is not None):
            self.__set_input(number=input_number, input=input, data=data)

        # Deal with the special case of options.
        if (option_name is not None) or (option is not None):
            self.__set_option(name=option_name, option=option, value=value, coerce_value=coerce_value)

        # Now update the model, persisting as needed.
        super(Process, self).update(**kwargs)

    @classmethod
    def __value_to_option(cls, value, data_type, validation):
        """
        Converts a Python option value into a string depending upon its data type and validation parameters.

        Args:
            value: The value to convert.
            data_type: The option data type.
            validation: The optional validation for the option.

        Returns:
            The correct string representation of the option.
        """
        if value is None:
            return None
        elif isinstance(value, bool):
            return str(value).lower()
        if data_type == fusion_platform.DATA_TYPE_DATETIME:
            return datetime.strftime(value, cls.__extract_datetime_validation(validation))
        else:
            return str(value)

    def wait_for_next_execution(self):
        """
        Waits for the next execution of the process to start, according to its schedule. If executions are already started, this method will return immediately.
        Otherwise, this method will block until the next scheduled execution has started.

        Raises:
            RequestError: if any request fails.
            ModelError: if a model could not be loaded or validated from the Fusion Platform<sup>&reg;</sup>.
        """
        # Wait until we find the next execution.
        while True:
            # Load in the most recent version of the model.
            self.get(organisation_id=self.organisation_id)

            # Get the most recent execution for the process. This assumes that the executions are returned with the most recent first.
            self._logger.debug('checking for next execution')
            executions = ProcessExecution._models_from_api_path(self._session, self._get_path(self.__class__._PATH_EXECUTIONS), items_per_request=1, reverse=True)
            execution = next(executions, None)

            # Ignore any execution older than when the next execution is expected. This assumes that the process repeat start is maintained correctly, and that it
            # is set after any corresponding executions have been created for the current repeat start.
            execution = None if (execution is not None) and (execution.created_at < self.repeat_start) else execution

            # Stop if we have an execution which is beyond the repeat start date.
            if execution is not None:
                self._logger.debug('execution %s found', execution.id)
                break

            # If we have no recent executions, and longer than the allowed period has elapsed since the next execution was meant to start, then raise an exception.
            if (execution is None) and (self.repeat_start + timedelta(seconds=Process._EXECUTE_WAIT_TOLERANCE)) < datetime.now(timezone.utc):
                raise ModelError(i18n.t('models.process.execution_should_have_started'))

            # We are waiting, so block for a short while.
            sleep(self.__class__._API_UPDATE_WAIT_PERIOD)
